"""
Functions for finding and reading files
"""

import os
import datetime
import typing
import h5py
import hdfmap
import numpy as np
from hdfmap.eval_functions import dataset2str, dataset2data
from imageio.v2 import imread

from mmg_toolbox.utils.misc_functions import consolidate_numeric_strings, regex_number


def list_files(folder_directory: str, extension='.nxs') -> list[str]:
    """Return list of files in directory with extension, returning list of full file paths ordered by modified time"""
    # return [os.path.join(folder_directory, file) for file in os.listdir(folder_directory) if file.endswith(extension)]
    try:
        return sorted(
            (file.path for file in os.scandir(folder_directory) if file.is_file() and file.name.endswith(extension)),
            key=lambda x: os.path.getmtime(x)
        )
    except (FileNotFoundError, PermissionError, OSError):
        return []


def display_timestamp(timestamp: float) -> str:
    """Return string representation of file timestamp"""
    return datetime.datetime.fromtimestamp(timestamp).strftime('%a %d-%b-%Y %H:%M')


def list_path_time(directory: str) -> list[tuple[str, float]]:
    """
    Return list of folders in diectory, along with modified time
        [(path, modified_time(s), nfiles), ...] = list_path_time_files('/folder/path', '.nxs')
    :param directory: directory to look in
    :return: [(path, timestamp), ...]
    """
    folders = [('.', os.stat(directory).st_mtime)]
    for f in os.scandir(directory):
        if f.is_dir():
            try:
                folders.append((f.path, f.stat().st_mtime))
            except PermissionError or FileNotFoundError:
                pass
    return sorted(folders, key=lambda x: x[0])


def list_folder_file_names(directory: str) -> tuple[list[str], list[str]]:
    """
    Return list of subdirectory and file names of given folder
    :param directory: directory to look in
    :return: folder_names, file_names
    """
    folders = []
    files = []
    for f in os.scandir(directory):
        if f.is_dir():
            try:
                folders.append(f.name)
            except PermissionError or FileNotFoundError:
                pass
        elif f.is_file():
            try:
                files.append(f.name)
            except PermissionError or FileNotFoundError:
                pass
    return folders, files


def folder_summary(directory: str) -> str:
    """Generate summary of folder"""
    folder_names, file_names = list_folder_file_names(directory)

    # subdirectories
    if len(folder_names) > 10:
        folder_names = consolidate_numeric_strings(*folder_names)
    if len(folder_names) > 30:
        subdirs_str = ""
    else:
        subdirs_str = '\n'.join(
            f"  {name}" for name in folder_names
        ) + '\n'
    # files
    files_str = '\n'.join(consolidate_numeric_strings(*file_names))
    file_ext = [os.path.splitext(file) for file in file_names]
    all_ext = {ext for name, ext in file_ext}
    file_types = {
        ext: (lst := [n for n, e in file_ext if e == ext], len(lst))
        for ext in all_ext
    }
    file_type_str = '\n'.join(
        f"  {ext}: {n}" for ext, (lst, n) in file_types.items()
    )
    summary = (
        f"Folder: {os.path.abspath(directory)}\n" +
        f"Modified: {display_timestamp(os.stat(directory).st_mtime)}\n\n" +
        f"Sub-Directories: {len(folder_names)}\n{subdirs_str}" +
        f"\nFiles: {len(file_names)}\n{files_str}\n" +
        f"\nFile-types:\n{file_type_str}"
    )
    return summary


def folder_summary_line(directory: str, extention='.nxs') -> str:
    """Generate summary of folder on a single line"""
    files = list_files(directory, extension=extention)
    ini = display_timestamp(os.path.getmtime(next(iter(files), directory)))
    fnl = display_timestamp(os.path.getmtime(next(reversed(files), directory)))
    return f"Files: {len(files)}, {ini} -> {fnl}"


def get_hdf_value(hdf_filename: str, hdf_address: str, default_value: typing.Any = '') -> typing.Any:
    """
    Open HDF file and return value from single dataset
    :param hdf_filename: str filename of hdf file
    :param hdf_address: str hdf address specifier of dataset
    :param default_value: Any - returned value if hdf_address is not available in file
    :return [dataset is array]: str "{type} {shape}"
    :return [dataset is not array]: output of dataset[()]
    :return [dataset doesn't exist]: default_value
    """
    try:
        with hdfmap.load_hdf(hdf_filename) as hdf:
            dataset = hdf.get(hdf_address)
            if isinstance(dataset, h5py.Dataset):
                return dataset2data(dataset)
            return default_value
    except Exception:
        return default_value


def get_hdf_string(hdf_filename: str, hdf_address: str, default_value: str = '') -> str:
    """
    Open HDF file and return value from single dataset
    :param hdf_filename: str filename of hdf file
    :param hdf_address: str hdf address specifier of dataset
    :param default_value: Any - returned value if hdf_address is not available in file
    :return [dataset is array]: str "{type} {shape}"
    :return [dataset is not array]: str output of dataset[()]
    :return [dataset doesn't exist]: default_value
    """
    try:
        with hdfmap.load_hdf(hdf_filename) as hdf:
            dataset = hdf.get(hdf_address)
            if isinstance(dataset, h5py.Dataset):
                return dataset2str(dataset)
            return default_value
    except Exception:
        return default_value


def hdfobj_string(hdf_filename: str, hdf_address: str) -> str:
    """Generate string describing object in hdf file"""
    with hdfmap.load_hdf(hdf_filename) as hdf:
        obj = hdf.get(hdf_address)
        if not obj:
            return ''
        try:
            link = repr(hdf.get(hdf_address, getlink=True))
        except RuntimeError:
            link = 'No link'
        myclass = hdf.get(hdf_address, getclass=True)
        out = f"{obj.name}\n"
        out += f"{repr(obj)}\n"
        out += f"{link}\n"
        out += f"{repr(myclass)}\n"
        out += '\nattrs:\n'
        out += '\n'.join([f"{key}: {obj.attrs[key]}" for key in obj.attrs])
        if isinstance(obj, h5py.Dataset):
            out += '\n\n--- Data ---\n'
            out += f"Shape: {obj.shape}\nSize: {obj.size}\nValues:\n"
            if obj.size > 1000:
                out += '---To large to view---'
            else:
                out += str(obj[()])
    return out


def read_tiff(image_filename: str) -> np.ndarray:
    """Read a tiff image, returning numpy array"""
    image = imread(image_filename)
    return np.array(image)


def get_scan_number(filename: str) -> int:
    """Return scan number from scan filename"""
    filename = os.path.basename(filename)
    match = regex_number.search(filename)
    if match:
        return int(match[0])
    return 0


def replace_scan_number(filename: str, new_number: int) -> str:
    """Replace scan number in filename"""
    path, filename = os.path.split(filename)
    new_filename = regex_number.sub(str(new_number), filename)
    return os.path.join(path, new_filename)



"""
Read measurement ASCII ####.dat files
Also known as SRS files
"""

import numpy as np

from mmg_toolbox.utils.misc_functions import DataHolder, data_holder


def read_dat_file(filename: str) -> DataHolder:
    """
    Reads #####.dat files from instrument, returns class instance containing all data
    Input:
      filename = string filename of data file
    Output:
      d = dict with attional attributes:
         d.*item* - returns scanned item as 1D array
         d.metadata - class containing all metadata from datafile
         d.metadata.*item* - returns value of metadata item
         d.keys() - returns all parameter names
         d.values() - returns all parameter values
         d.items() - returns parameter (name,value) tuples
    """
    with open(filename, 'r') as f:
        lines = f.readlines()

    # Read metadata
    meta = {}
    lineno = 0
    for ln in lines:
        lineno += 1
        if '&END' in ln: break
        ln = ln.strip(' ,\n')
        neq = ln.count('=')
        if neq == 1:
            'e.g. cmd = "scan x 1 10 1"'
            inlines = [ln]
        elif neq > 1:
            'e.g. SRSRUN=571664,SRSDAT=201624,SRSTIM=183757'
            inlines = ln.split(',')
        else:
            'e.g. <MetaDataAtStart>'
            continue

        for inln in inlines:
            vals = inln.split('=')
            if len(vals) != 2: continue  # skip any lines with more than 1 '='
            try:
                meta[vals[0]] = eval(vals[1])
            except SyntaxError: # catch strings without quotations
                meta[vals[0]] = vals[1]

    # Read Scanned data
    # previous loop ended at &END, now starting on list of names
    names = lines[lineno].split()
    # Load 2D arrays of scanned values
    vals = np.loadtxt(lines[lineno + 1:], ndmin=2)
    # Assign arrays to a dictionary
    scanned = {
        name: value for name, value in zip(names, vals.T)
    }

    # Convert to class instance
    return data_holder(scanned, meta)



import tkinter as tk

from mmg_toolbox.tkguis.misc.functions import topmenu
from mmg_toolbox.tkguis.misc.styles import RootWithStyle, create_root


def create_file_browser(parent: tk.Misc | None = None, initial_directory: str | None = None) -> RootWithStyle:
    """
    File Browser - Browse directories and open NeXus files
    """
    from ..widgets.folder_treeview import FolderTreeViewFrame

    root = create_root(parent=parent, window_title='File Browser')
    topmenu(root, {}, add_themes=True, add_about=True)
    FolderTreeViewFrame('Any', root, initial_directory)
    if parent is None:
        root.mainloop()
    return root


def create_nexus_file_browser(parent: tk.Misc | None = None, initial_directory: str | None = None,
                              hdf_path: str = '/entry/scan_command') -> RootWithStyle:
    """
    File Browser - Browse directories and open NeXus files
    """
    from ..widgets.folder_treeview import NexusFolderTreeViewFrame

    root = create_root(parent=parent, window_title='NeXus File Browser')
    topmenu(root, {}, add_themes=False, add_about=True)
    NexusFolderTreeViewFrame(root, initial_directory, hdf_path)
    if parent is None:
        root.mainloop()
    return root


def create_jupyter_browser(parent: tk.Misc | None = None, initial_directory: str | None = None) -> RootWithStyle:
    """
    File Browser - Browse directories and open NeXus files
    """
    from ..widgets.folder_treeview import JupyterFolderTreeViewFrame

    root = create_root(parent=parent, window_title='Jupyter Notebook Browser')
    topmenu(root, {}, add_themes=True, add_about=True)
    JupyterFolderTreeViewFrame(root, initial_directory)
    if parent is None:
        root.mainloop()
    return root



"""
Set of functions for analysing x-ray absorption spectra
"""

import json

import os
import numpy as np
from lmfit.model import ModelResult
from lmfit.models import LinearModel, QuadraticModel, ExponentialModel, StepModel, PolynomialModel

# trapz replaced by trapezoid in Numpy 2.0
try:
    from numpy import trapezoid as trapz
except ImportError:
    from numpy import trapz


EDGE_FILE = os.path.join(os.path.dirname(__file__), '..', 'data', 'xray_edges.json')
SEARCH_EDGES = ('L3', 'L2')


def load_edge_energies(edges=SEARCH_EDGES) -> tuple[np.ndarray, np.ndarray]:
    """
    return arrays of energies and labels for x-ray absorption edges
    :param edges: if not None, only return energies for these edges, e.g. ('L3', 'L2')
    :return: energies[ndarray], labels[ndarray]
    """
    with open(EDGE_FILE, 'r') as infile:
        edge_dict = json.load(infile)
    edge_energies = {
        label: edge['energy'] for label, edge in edge_dict.items()
        if (edge['edge'] in edges if edges else True)
    }
    energies = np.array(list(edge_energies.values()))
    labels = np.array(list(edge_energies.keys()))
    idx = np.argsort(energies)
    return energies[idx], labels[idx]


def xray_edges_in_range(min_energy_ev: float, max_energy_ev: float | None = None,
                        energy_range_ev: float = 10., search_edges: None | tuple[str] = SEARCH_EDGES) -> list[tuple[str, float]]:
    """
    Return all x-ray absorption edges within the range
    :param min_energy_ev: energy to find x-ray absorption edges within
    :param max_energy_ev: energy to find x-ray absorption edges within
    :param energy_range_ev: energy to find x-ray absorption edges within
    :param search_edges: if not None, only return energies for these edges, e.g. ('L3', 'L2')
    :return: list[(edge_label[str], energy[float])]
    """
    if max_energy_ev is None:
        min_energy_ev = min_energy_ev - energy_range_ev / 2
        max_energy_ev = min_energy_ev + energy_range_ev / 2
    energies, labels = load_edge_energies(search_edges)
    idx = (energies > min_energy_ev) * (energies < max_energy_ev)
    return [(str(labels[ii]), float(energies[ii])) for ii in np.flatnonzero(idx)]


def energy_range_edge_label(min_energy_ev: float, max_energy_ev: float | None = None,
                            energy_range_ev: float = 10., search_edges: tuple[str] = SEARCH_EDGES) -> tuple[str, str]:
    """
    Return mode string for x-ray absorption edges in energy range
      raises ValueError is no edges are found or if multiple non-equivalent edges are found

    :param min_energy_ev: energy to find x-ray absorption edges within
    :param max_energy_ev: energy to find x-ray absorption edges within
    :param energy_range_ev: energy to find x-ray absorption edges within
    :param search_edges: if not None, only return energies for these edges, e.g. ('L3', 'L2')
    :return: element, mode strings, e.g. 'Mn', 'L2, L3'
    """
    edges = xray_edges_in_range(min_energy_ev, max_energy_ev, energy_range_ev, search_edges)
    if len(edges) == 1:
        label = edges[0][0]
        element, edge = label.split()
        return element, edge
    if len(edges) == 2:
        label1 = edges[0][0]
        label2 = edges[1][0]
        element1, edge1 = label1.split()
        element2, edge2 = label2.split()
        if element1 != element2:
            raise ValueError(f"xray absorption edges of multiple edges present: {label1}, {label2}")
        return element1, f"{edge1}, {edge2}"
    raise ValueError(f"xray absorption edge not found: {edges} edges at energy {min_energy_ev} eV")


def average_energy_scans(*args: np.ndarray):
    """Return the minimum range covered by all input arguments"""
    min_energy = np.max([np.min(en) for en in args])
    max_energy = np.min([np.max(en) for en in args])
    min_step = np.min([np.min(np.abs(np.diff(en))) for en in args])
    return np.arange(min_energy, max_energy + min_step, min_step)


def average_energy_spectra(energy, *args: tuple[np.ndarray, np.ndarray]):
    """
    Average energy spectra, interpolating at given energy

    E.G.
        energy = average_energy_scans(en1, en2)
        signal = combine_energy_scans(energy, (en1, sig1), (en2, sig2))

    :param energy: (n*1) array of energy values, in eV
    :param args: (mes_energy, mes_signal): pair of (m*1) arrays for energy and measurement raw_signals
    :returns signal: (n*1) array of averaged signal values at points in energy
    """
    data = np.zeros([len(args), len(energy)])
    for n, (en, dat) in enumerate(args):
        data[n, :] = np.interp(energy, en, dat)
    return data.mean(axis=0)


def preedge_signal(energy, signal, ev_from_start=5.) -> float:
    """Return pre-edge signal"""
    return np.mean(signal[energy < np.min(energy) + ev_from_start])


def postedge_signal(energy, signal, ev_from_end=5.) -> float:
    """Return post-edge signal"""
    return np.mean(signal[energy > np.max(energy) - ev_from_end])


def signal_jump(energy, signal, ev_from_start=5., ev_from_end=None) -> float:
    """Return signal jump from start to end"""
    ev_from_end = ev_from_end or ev_from_start
    ini_signal = preedge_signal(energy, signal, ev_from_start)
    fnl_signal = postedge_signal(energy, signal, ev_from_end)
    return fnl_signal - ini_signal


"""
--- BACKGROUND FUNCTIONS ---
All background functions have inputs:
    energy: ndarray[n], signal: ndarray[n], **kwargs
All background functions have outputs:
    bkg: nparray[n], norm: float, None | lmfit.ModelResult
"""


def subtract_flat_background(energy, signal, ev_from_start=5.) -> tuple[np.ndarray, float, None]:
    """Subtract flat background"""
    bkg = preedge_signal(energy, signal, ev_from_start)
    return bkg * np.ones_like(signal), 1, None


def normalise_background(energy, signal, ev_from_start=5.) -> tuple[np.ndarray, float, None]:
    """Normalise background to one"""
    bkg = preedge_signal(energy, signal, ev_from_start)
    return np.zeros_like(signal), float(bkg), None


def fit_linear_background(energy, signal, ev_from_start=5.) -> tuple[np.ndarray, float, ModelResult]:
    """Use lmfit to determine sloping background"""
    model = LinearModel(prefix='bkg_')
    region = energy < np.min(energy) + ev_from_start
    en_region = energy[region]
    sig_region = signal[region]
    pars = model.guess(sig_region, x=en_region)
    fit_output = model.fit(sig_region, pars, x=en_region)
    bkg = fit_output.eval(x=energy)
    return bkg, 1, fit_output


def fit_curve_background(energy, signal, ev_from_start=5.) -> tuple[np.ndarray, float, ModelResult]:
    """Use lmfit to determine sloping background"""
    model = QuadraticModel(prefix='bkg_')
    # region = (energy < np.min(energy) + ev_from_start) + (energy > np.max(energy) - ev_from_start)
    region = energy < np.min(energy) + ev_from_start
    en_region = energy[region]
    sig_region = signal[region]
    pars = model.guess(sig_region, x=en_region)
    fit_output = model.fit(sig_region, pars, x=en_region)
    bkg = fit_output.eval(x=energy)
    return bkg, 1, fit_output


def fit_exp_background(energy, signal, ev_from_start=5.) -> tuple[np.ndarray, float, ModelResult]:
    """Use lmfit to determine sloping background"""
    model = ExponentialModel(prefix='bkg_')
    # region = (energy < np.min(energy) + ev_from_start) + (energy > np.max(energy) - ev_from_start)
    region = energy < np.min(energy) + ev_from_start
    en_region = energy[region]
    sig_region = signal[region]
    pars = model.guess(sig_region, x=en_region)
    fit_output = model.fit(sig_region, pars, x=en_region)
    # print('exp background\n:', fit_output.fit_report())
    bkg = fit_output.eval(x=energy)
    return bkg, 1, fit_output


def fit_step_background(energy, signal, ev_from_start=5.)  -> tuple[np.ndarray, float, ModelResult]:  # good?
    """Use lmfit to detemine edge background"""
    model = LinearModel(prefix='bkg_') + StepModel(form='arctan', prefix='edge_')
    region = (energy < np.min(energy) + ev_from_start) + (energy > np.max(energy) - ev_from_start)
    en_region = energy[region]
    sig_region = signal[region]

    guess_jump = signal_jump(energy, signal)
    pars = model.make_params(
        bkg_slope=0.0,
        bkg_intercept=np.min(sig_region),
        edge_amplitude=guess_jump,
        edge_center=np.mean(energy),
        edge_sigma=1.0,
    )
    # bkg_ini = model.eval(pars, x=energy)
    fit_output = model.fit(sig_region, pars, x=en_region)
    bkg = fit_output.eval(x=energy)
    step = fit_output.params['edge_amplitude']
    return bkg, step, fit_output


def fit_double_edge_step_background(energy, signal, l3_energy, l2_energy, peak_width_ev=5.) -> tuple[np.ndarray, float, ModelResult]:
    """Use lmfit to determine sloping background"""
    model = StepModel(form='arctan', prefix='l3_') + StepModel(form='arctan', prefix='l2_')  # form='linear'
    region = (
            (energy < l3_energy - peak_width_ev / 2) +
            np.logical_and(energy > l3_energy + peak_width_ev / 2, energy < l2_energy - peak_width_ev / 2) +
            (energy > l2_energy + peak_width_ev / 2)
    )
    en_region = energy[region]
    sig_region = signal[region]

    guess_jump = signal_jump(energy, signal)
    pars = model.make_params(
        l3_amplitude=0.667 * guess_jump,
        l3_center=l3_energy,
        l3_sigma=2,
        l2_amplitude=0.333 * guess_jump,
        l2_center=l2_energy,
        l2_sigma=2,
    )
    pars['l3_center'].set(min=l3_energy - 1.0, max=l3_energy + 1.0)
    pars['l2_center'].set(min=l2_energy - 1.0, max=l2_energy + 1.0)
    pars['l3_sigma'].set(min=1, max=5)
    # pars['l2_sigma'].set(min=1, max=5)
    pars['l2_sigma'].set(expr='l3_sigma')
    pars['l3_amplitude'].set(min=0.6 * guess_jump, max=0.75 * guess_jump)
    pars['l2_amplitude'].set(expr=f"{guess_jump}-l3_amplitude")
    # bkg_ini = model.eval(pars, x=energy)
    fit_output = model.fit(sig_region, pars, x=en_region)
    bkg = fit_output.eval(x=energy)
    step = fit_output.params['l3_amplitude'] + fit_output.params['l2_amplitude']
    return bkg, step, fit_output


def fit_spectra_background(energy: np.ndarray, signal: np.ndarray, *step_energies: float, peak_width_ev=5.) -> tuple[np.ndarray, float, ModelResult]:
    """
    Generic fit of spectra background using an order-2 polynomial and n-edges, fitted to region with peaks removed

    The returned ModelResult object has the following attributes:
        params['bkg_0']  flat background
        params['bkg_1']  sloping background
        params['bkg_2']  curved background
        params['edgeN_center']  step N centre [in eV]
        params['edgeN_amplitude']  step N height
        params['edgeN_sigma']  step N width [in eV]

    Parameters
    :energy: ndarray[n] of spectra energy in eV
    :signal: ndarray[n] of spectra signal
    :step_energies: list of absorption energy steps, in eV
    :peak_width_ev: float width of absorption peak in eV
    :return: background[ndarray], jump[float], lmfit.ModelResult
    """
    model = PolynomialModel(degree=2, prefix='bkg_')
    region = np.ones_like(energy, dtype=bool)
    for n, edge in enumerate(step_energies):
        model += StepModel(form='arctan', prefix=f'edge{n+1}_')
        region[np.abs(energy - edge) < peak_width_ev] = 0
    en_region = energy[region]
    sig_region = signal[region]

    guess_jump = signal_jump(energy, signal)
    pars = model.make_params(
        bkg_c0=np.min(sig_region),
        bkg_c1=0,
        bkg_c2=0
    )
    for n, edge in enumerate(step_energies):
        pars[f'edge{n+1}_center'].set(value=edge, min=edge - 2.0, max=edge + 2.0)
        pars[f'edge{n+1}_sigma'].set(value=2, min=1, max=3)
        if n > 0:
            pars[f'edge{n + 1}_sigma'].set(expr='edge1_sigma')
        edge_jump = guess_jump * (len(step_energies) - n) / (len(step_energies) + 1)
        pars[f'edge{n + 1}_amplitude'].set(value=edge_jump, min=0.8 * edge_jump, max=1.2 * edge_jump)
    # bkg_ini = model.eval(pars, x=energy)
    fit_output = model.fit(sig_region, pars, x=en_region)
    bkg = fit_output.eval(x=energy)
    jump = 0
    for n, edge in enumerate(step_energies):
        jump += fit_output.params[f'edge{n + 1}_amplitude']
    return bkg, jump, fit_output


def fit_spectra_exp_background(energy: np.ndarray, signal: np.ndarray, *step_energies: float, peak_width_ev=5.) -> tuple[np.ndarray, float, ModelResult]:
    """
    Generic fit of spectra background using an exponential and n-edges, fitted to region with peaks removed

    The returned ModelResult object has the following attributes:
        params['bkg_0']  flat background
        params['bkg_1']  sloping background
        params['bkg_2']  curved background
        params['edgeN_center']  step N centre [in eV]
        params['edgeN_amplitude']  step N height
        params['edgeN_sigma']  step N width [in eV]

    Parameters
    :energy: ndarray[n] of spectra energy in eV
    :signal: ndarray[n] of spectra signal
    :step_energies: list of absorption energy steps, in eV
    :peak_width_ev: float width of absorption peak in eV
    :return: background[ndarray], jump[float],  lmfit.ModelResult
    """
    model = ExponentialModel(prefix='bkg_')
    region = np.ones_like(energy, dtype=bool)
    for n, edge in enumerate(step_energies):
        model += StepModel(form='arctan', prefix=f'edge{n+1}_')
        region[np.abs(energy - edge) < peak_width_ev] = 0
    en_region = energy[region]
    sig_region = signal[region]

    guess_jump = signal_jump(energy, signal)
    pars = model.make_params(
        bkg_amplitude=np.max(sig_region),
        bkg_decay=100.0,
    )
    for n, edge in enumerate(step_energies):
        pars[f'edge{n+1}_center'].set(value=edge, min=edge - 2.0, max=edge + 2.0)
        pars[f'edge{n+1}_sigma'].set(value=2, min=1, max=3)
        if n > 0:
            pars[f'edge{n + 1}_sigma'].set(expr='edge1_sigma')
        edge_jump = guess_jump * (len(step_energies) - n) / (len(step_energies) + 1)
        pars[f'edge{n + 1}_amplitude'].set(value=edge_jump, min=0.8 * edge_jump, max=1.2 * edge_jump)
    # bkg_ini = model.eval(pars, x=energy)
    fit_output = model.fit(sig_region, pars, x=en_region)
    bkg = fit_output.eval(x=energy)
    jump = 0
    for n, edge in enumerate(step_energies):
        jump += fit_output.params[f'edge{n + 1}_amplitude']
    return bkg, jump, fit_output


"""
--- SUM RULES ---
"""


def default_n_holes(element: str) -> float:
    """
    Return the default number of holes for a given element
    """
    elements = {
        'Cu': 1,
        'Ni': 2,
        'Co': 3,
        'Fe': 4,
        'Mn': 5,
        'Cr': 6,
        'V':  7,
        'Ti': 8,
        'Sc': 9,
    }
    if element in elements:
        return elements[element]
    raise KeyError(f'unknown number of holes for {element}')


def orbital_angular_momentum(energy: np.ndarray, average: np.ndarray,
                             difference: np.ndarray, nholes: float) -> float:
    """
    Calculate the sum rule for the angular momentum of the spectra
    using the formula:
    L = -2 * nholes * int[spectra d energy] / sum(spectra)

    :param energy: Energy axis of the spectra
    :param average: average XAS spectra (left + right) for both polarisations
    :param difference: difference XAS spectra (right - left) for both polarisations
    :param nholes: Number of holes in the system
    :return: Angular momentum of the spectra
    """
    if len(energy) != len(average) or len(energy) != len(difference):
        raise ValueError(f"Energy and spectra must have the same length: {len(energy)} != {len(average)}")
    if nholes <= 0:
        raise ValueError(f"Number of holes must be greater than 0: {nholes}")

    # total intensity
    tot = trapz(average, energy)

    # Calculate the sum rule for the angular momentum
    L = -2 * nholes * trapz(difference, energy) / tot
    return L


def spin_angular_momentum(energy: np.ndarray, average: np.ndarray,
                          difference: np.ndarray, nholes: float,
                          split_energy: int | None = None, dipole_term: float = 0) -> float:
    """
    Calculate the sum rule for the spin angular momentum of the spectra
    using the formula:
    S = -2 * nholes * int[spectra d energy] / sum(spectra)

    :param energy: Energy axis of the spectra
    :param average: average XAS spectra (left + right) for both polarisations
    :param difference: difference XAS spectra (right - left) for both polarisations
    :param nholes: Number of holes in the system
    :param split_energy: energy to split the spectra between L3 and L2 (or None to use the middle of the spectra)
    :param dipole_term: magnetic dopole term (T_z), defaults to 0 for effective spin
    :return: Spin angular momentum of the spectra
    """
    if len(energy) != len(average) or len(energy) != len(difference):
        raise ValueError(f"Energy and spectra must have the same length: {len(energy)} != {len(average)}")
    if nholes <= 0:
        raise ValueError(f"Number of holes must be greater than 0: {nholes}")
    if split_energy is None:
        split_energy = (energy[0] + energy[-1]) / 2

    # total intensity
    tot = trapz(average, energy)

    # Calculate the sum rule for the spin angular momentum
    split_index = np.argmin(np.abs(energy - split_energy))
    l3_energy = energy[split_index:]  # L3 edge at lower energy
    l3_difference = difference[split_index:]
    l3_integral = trapz(l3_difference, l3_energy)
    l2_energy = energy[:split_index]
    l2_difference = difference[:split_index]
    l2_integral = trapz(l2_difference, l2_energy)
    S_eff = (3 / 2) * nholes * (l3_integral - 2 * l2_integral) / tot
    S = S_eff - dipole_term
    return S


def magnetic_moment(orbital: float, spin: float) -> float:
    """
    Calculate the magnetic moment of the system using the formula:
    M = -g * (L + 2 * S)  WHERE DOES THIS COME FROM?

    :param orbital: Orbital angular momentum of the system
    :param spin: Spin angular momentum of the system
    :return: Magnetic moment of the system
    """
    print('magnetic moment is probably wrong!')
    g = 2.0  # Landé g-factor for free electron
    return -g * (orbital + 2 * spin)



"""
Main MMG Runscript
"""

import sys
from . import start_gui


if __name__ == '__main__':
    if 'gui' in sys.argv:
        start_gui()

    # Start up interactive console
    import sys
    import numpy as np
    import matplotlib.pyplot as plt
    import mmg_toolbox as mmg

    print(f"""
    {mmg.version_info()}
    By Dan Porter, Diamond Light Source Ltd.
    
    See help(mmg) for info, or mmg.start_gui() to get started!
    """)





"""
Convenience wrapper for ttk.Treeview
"""

import tkinter as tk
from tkinter import ttk

from mmg_toolbox.tkguis.misc.functions import post_right_click_menu


TreeViewColumn = tuple[str, str, int, bool, str | None]  # name, title, width, reverse, sort_col


def treeview_sort_column(treeview: ttk.Treeview, col: str, reverse: bool, sort_col: str | None = None):
    """
    Function to sort columns in ttk.Treeview,
        tree.heading("#0", command=lambda _col="#0": treeview_sort_column(tree, _col, False))
    :param treeview: ttk.Treeview instance
    :param col: str, column specifier for items to sort
    :param reverse: Bool, sort direction
    :param sort_col: str or None, sort alternative column
    :return:
    """
    if sort_col is None:
        sort_col = col
    
    c_item = lambda iid: treeview.item(iid)['text'] if col == '#0' else lambda iid: treeview.set(iid, col)
    # if col == "#0":
    #     def get_item(iid):
    #         return treeview.item(iid)['text']
    # else:
    #     def get_item(iid):
    #         return treeview.set(iid, col)

    items = [(c_item(iid), iid) for iid in treeview.get_children('')]
    items.sort(reverse=reverse)

    # rearrange items in sorted positions
    for index, (val, k) in enumerate(items):
        treeview.move(k, '', index)
        if treeview.item(k)['text'] == '..':  # keep at top of column
            treeview.move(k, '', 0)

    # reverse sort next time
    treeview.heading(sort_col, command=lambda _col=col: treeview_sort_column(treeview, _col, not reverse, sort_col))


class CanvasTreeview:
    """
    Treeview wrapper comprising a ttk.Treeview inside a canvas with scrollbars

        columns = [
            ('#0', "Number", 100, False, None),
            ('file', "Filename", 400, False, None),
        ]
        tv = CanvasTreeview(parent, *columns)

    *Note: the *name* parameter of the first column must be '#0'.

    :param root: parent tk Frame object
    :param columns: list of tuples where each tuple is ('name', 'Title', width, reverse, sort_col)
    :param width: width of widget, or None to fill and expand
    :param height: heigh of widget, or None to fill and expand
    """
    def __init__(self, root: tk.Misc, *columns: TreeViewColumn,
                 width: int | None = None, height: int | None = None,
                 pack: bool = True):
        self.root = root

        canvas = tk.Canvas(root)
        # fixed size of grid in canvas
        canvas.grid_rowconfigure(0, weight=1)
        canvas.grid_columnconfigure(0, weight=1)
        if width and height:
            canvas.configure(width=width, height=height)
            self.pack_treeview = lambda: canvas.pack()
        else:
            self.pack_treeview = lambda: canvas.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
        if pack:
            self.pack_treeview()
        canvas.grid_propagate(False)

        frm = ttk.Frame(canvas)
        frm.grid_rowconfigure(0, weight=1)
        frm.grid_columnconfigure(0, weight=1)

        tree = ttk.Treeview(frm, columns=[c[0] for c in columns[1:]])
        for c in columns:
            tree.column(c[0], stretch=False)

        var = ttk.Scrollbar(frm, orient="vertical", command=tree.yview)
        # var.pack(side=tk.RIGHT, fill=tk.Y)
        var.grid(column=1, row=0, sticky='ns')
        tree.configure(yscrollcommand=var.set)

        var = ttk.Scrollbar(frm, orient="horizontal", command=tree.xview)
        # var.pack(side=tk.BOTTOM, fill=tk.X)
        var.grid(column=0, row=1, sticky='ew')
        tree.configure(xscrollcommand=var.set)
        # tree.pack(side=tk.TOP)
        tree.grid(column=0, row=0, sticky='nsew')
        # place the frame inside the canvas as a window and make it resize with the canvas
        _canvas_window = canvas.create_window(0, 0, anchor='nw', window=frm)

        def _on_canvas_configure(event):
            # ensure the inner frame (and thus the tree) matches the canvas size
            canvas.itemconfigure(_canvas_window, width=event.width, height=event.height)

        canvas.bind('<Configure>', _on_canvas_configure)

        def tree_sort(col, reverse, sort_col=None):
            return lambda: treeview_sort_column(tree, col, reverse=reverse, sort_col=sort_col)

        for name, title, width, _reverse, _sort_col in columns:
            tree.heading(name, text=title, command=tree_sort(_sort_col or name, _reverse, name if _sort_col else None))
            tree.column(name, width=width, stretch=False)  # stretch stops columns from stretching when resized

        # Hide columns
        display_columns = [name for name, title, width, reverse, sort_col in columns[1:] if width > 0]
        tree.configure(displaycolumns=display_columns)

        self.titles = {
            name: title for name, title, width, reverse, sort_col in columns
        }
        self.getters = {
            '#0': lambda iid: self.tree.item(iid)['text'],
            **{
                name: lambda iid, _name=name: self.tree.set(iid, _name)
                for name, _, _, _, _ in columns[1:]
            }
        }

        self.columns = columns
        self.tree = tree
        tree.bind("<Button-3>", self.right_click_menu())

    def first_item(self):
        return next(iter(self.tree.get_children()))

    def get_row(self, iid: str | int) -> dict:
        """Return data from row"""
        return {name: getter(iid) for name, getter in self.getters.items()}

    def get_selected(self) -> list[dict]:
        """Return values of current selection"""
        return [self.get_row(iid) for iid in self.tree.selection()]

    def get_index(self):
        iid = next(iter(self.tree.selection()), next(iter(self.tree.get_children())))
        return self.tree.index(iid)

    def populate(self, **kwargs):
        pass

    def delete(self):
        self.tree.delete(*self.tree.get_children())

    def bind_select(self, function):
        self.tree.bind('<<TreeviewSelect>>', function)

    def bind_dbl_click(self, function):
        self.tree.bind("<Double-1>", function)

    def right_click_menu(self):
        """
        Create right-click context menu for hdf_tree objects
        :return: menu_popup function
        """

        def copy_fun(tree_getter):
            def fun():
                for iid in self.tree.selection():
                    self.root.master.clipboard_clear()
                    self.root.master.clipboard_append(tree_getter(iid))
            return fun

        m = tk.Menu(self.root, tearoff=0)
        header_name = self.tree.heading('#0', 'text')
        header_getter = lambda iid: self.tree.item(iid)['text']
        m.add_command(label="Copy " + header_name, command=copy_fun(header_getter))
        for name, title, width, reverse, sort_col in self.columns:
            getter = lambda iid: self.tree.set(iid, name)
            m.add_command(label="Copy " + title, command=copy_fun(getter))

        def menu_popup(event):
            # select item
            iid = self.tree.identify_row(event.y)
            if iid:
                self.tree.selection_set(iid)
                post_right_click_menu(m, event.x_root, event.y_root)

        return menu_popup



import tkinter as tk

from mmg_toolbox.tkguis.misc.config import get_config, C
from mmg_toolbox.tkguis.misc.styles import create_root, RootWithStyle


def list_scans(*file_list: str, parent: tk.Misc | None = None, config: dict | None = None,
               metadata_list: list[str] | None = None, button_name: str = 'Close') -> list[str]:
    """
    Create a window displaying a list of scans with metadata
    """
    from ..widgets.scan_selector import ScanViewer

    root = create_root(parent=parent, window_title='Select Files')
    config = get_config() if config is None else config
    if metadata_list:
        # replace metadata_list in config
        config = config.copy()
        config[C.metadata_list] = {name: f"{{{name}}}" for name in metadata_list}
    return ScanViewer(root, *file_list, config=config, button_name=button_name).show()


def create_range_selector(initial_folder: str | None = None,
                          parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    Create a range selector
    """
    from ..widgets.scan_range_selector import ScanRangeSelector

    root = create_root(parent=parent, window_title='NeXus Scan Viewer')
    config = get_config() if config is None else config

    ScanRangeSelector(root, initial_folder, config)

    if parent is None:
        root.mainloop()
    return root



"""
Handle small screens by adjusting size of elements
"""

import tkinter as tk
from .config import (C, TEXT_HEIGHT, TEXT_HEIGHT_SMALL, TEXT_WIDTH,
                     FIGURE_SIZE, IMAGE_SIZE, FIGURE_DPI, MAX_PLOT_SCREEN_PERCENTAGE, SMALL_SCREEN_HEIGHT)


def get_screen_size_inches(root: tk.Misc, dpi: int = 60) -> tuple[float, float]:
    """Return the screen size in inches: [width, height]"""
    width, height = root.winfo_screenwidth(), root.winfo_screenheight()
    return width / dpi, height / dpi


def get_text_size(root: tk.Misc, config: dict | None = None) -> tuple[int, int]:
    """Return textbox size adjusted for screen size in [characters, lines]"""
    width, height = root.winfo_screenwidth(), root.winfo_screenheight()
    config = config or {}
    if height <= config.get(C.small_screen_height, SMALL_SCREEN_HEIGHT):
        print(f"Small screen size, reducing Text Height to {TEXT_HEIGHT_SMALL}")
        characters, lines = config.get(C.text_size_small, (TEXT_WIDTH, TEXT_HEIGHT_SMALL))
    else:
        characters, lines = config.get(C.text_size, (TEXT_WIDTH, TEXT_HEIGHT))
    return characters, lines


def get_figure_size(root: tk.Misc, config: dict | None = None, config_label: str = C.plot_size) -> tuple[int, int]:
    """Return figure size adjusted for screen size in [width, height] inches"""
    config = config or {}

    fig_width, fig_height = config.get(config_label, FIGURE_SIZE)
    width_inches, height_inches = get_screen_size_inches(root, config.get(C.plot_dpi, FIGURE_DPI))
    wid_perc = 100 * (fig_width / width_inches)
    hgt_perc = 100 * (fig_height / height_inches)
    max_wid, max_hgt = config.get(C.plot_max_percent, MAX_PLOT_SCREEN_PERCENTAGE)

    if hgt_perc > max_hgt:
        print(f"Figure size as % of screen: {width_inches:.2f}x{height_inches:.2f}\" = {wid_perc:.0f}x{hgt_perc:.0f}%")
        print(f"Reducing to {max_hgt}% height")
        fig_height = max_hgt * height_inches / 100.
    return fig_width, fig_height


def update_config(root: tk.Misc, config: dict) -> None:
    """Update config for screen size"""
    width, height = root.winfo_screenwidth(), root.winfo_screenheight()
    small_height = config.get(C.small_screen_height, SMALL_SCREEN_HEIGHT)
    scale = 1.0 if height > small_height else 0.6
    new = {
        C.text_size: (TEXT_WIDTH, TEXT_HEIGHT_SMALL if height > small_height else TEXT_HEIGHT),
        C.plot_size: (FIGURE_SIZE[0], FIGURE_SIZE[1] * scale),
        C.image_size: (IMAGE_SIZE[0], IMAGE_SIZE[1] * scale),
    }
    config.update(new)


"""
tkinter widget to find scans
"""

import tkinter as tk
from tkinter import ttk
import hdfmap

from mmg_toolbox.utils.experiment import Experiment
from ..misc.config import C
from ..misc.logging import create_logger
from ..apps.namespace_select import create_metadata_selector

logger = create_logger(__file__)


class FindScans:
    """
    tkinter frame to find scans
    """
    def __init__(self, root: tk.Misc, exp_folder: str, config: dict, scan_file: str | None = None,
                 metadata_list: list[str] | None = None, close_fun=None):
        self.root = root
        self.close_fun = root.destroy if close_fun is None else close_fun
        self.exp = Experiment(exp_folder, instrument=config.get(C.beamline, None))
        self.scan_file = scan_file or self.exp.get_scan_filename(-1)
        self.hdf_map = hdfmap.create_nexus_map(self.scan_file)
        self.scan_numbers = []
        self.vars: list[tuple[tk.StringVar, tk.StringVar, tk.StringVar, tk.StringVar]] = []
        self.return_on_first = False
        metadata_list = metadata_list or []

        window = tk.Frame(self.root)
        window.pack(fill=tk.BOTH, expand=tk.YES, padx=2, pady=2)

        ttk.Label(window, text='Find Scans', style='subtitle.TLabel').pack(side=tk.TOP, pady=5)
        self.var_sec = ttk.Frame(window, relief=tk.RIDGE, borderwidth=2)
        self.var_sec.pack(side=tk.TOP, fill=tk.BOTH, padx=2, pady=2)

        line = ttk.Frame(self.var_sec)
        line.pack(side=tk.TOP, fill=tk.X, padx=2, pady=3)
        ttk.Label(line, text='Name / expression', width=20).pack(side=tk.LEFT, padx=5)
        ttk.Label(line, text='Value', width=10).pack(side=tk.LEFT, padx=2)
        ttk.Label(line, text='Tolerance', width=10).pack(side=tk.LEFT, padx=2)

        self.add_vars(*metadata_list)

        sec = ttk.Frame(self.var_sec)
        sec.pack(side=tk.BOTTOM, fill=tk.X, padx=2, pady=5)
        ttk.Button(sec, text='Add', command=self.add_vars).pack()

        sec = ttk.Frame(window)
        sec.pack(side=tk.BOTTOM, fill=tk.X, padx=2, pady=5)
        ttk.Button(sec, text='Find Scans', command=self.find_scans).pack(side=tk.LEFT, padx=3)
        ttk.Button(sec, text='Close', command=self.close_fun).pack(side=tk.LEFT, padx=3)

    def add_vars(self, *metadata_names: str):
        metadata_names = metadata_names + ('', )
        for name in metadata_names:
            var_name = tk.StringVar(self.root, name)
            var_lab = tk.StringVar(self.root, '')
            var_val = tk.StringVar(self.root, '')
            var_tol = tk.StringVar(self.root, '')
            self.vars.append((var_name, var_lab, var_val, var_tol))
            self.add_var_line(var_name, var_lab, var_val, var_tol)

    def add_var_line(self, var_name: tk.StringVar, var_lab: tk.StringVar,
                     var_val: tk.StringVar, var_tol: tk.StringVar):
        def update_val(_event=None):
            if var_name.get():
                val = self.hdf_map.eval(self.hdf_map.load_hdf(), var_name.get())
                var_val.set(val)
                if isinstance(val, str):
                    var_lab.set('contains')
                    var_tol.set('--')
                else:
                    var_lab.set('~=')
                    var_tol.set('1.0')

        def select():
            metadata = create_metadata_selector(self.hdf_map, self.root)
            if metadata:
                var_name.set(metadata[0])
                update_val()

        def remove():
            var_name.set('')
            var_val.set('')
            var_tol.set('')

        line = ttk.Frame(self.var_sec)
        line.pack(side=tk.TOP, fill=tk.X, padx=2, pady=3)
        ttk.Button(line, text=':', width=1, command=select).pack(side=tk.LEFT)
        var = ttk.Entry(line, textvariable=var_name, width=20)
        var.pack(side=tk.LEFT, padx=5)
        var.bind('<Return>', update_val)
        ttk.Label(line, textvariable=var_lab, width=10, anchor=tk.E).pack(side=tk.LEFT, padx=2)
        ttk.Entry(line, textvariable=var_val, width=10).pack(side=tk.LEFT, padx=2)
        ttk.Label(line, text='+/-').pack(side=tk.LEFT)
        ttk.Entry(line, textvariable=var_tol, width=10).pack(side=tk.LEFT, padx=2)
        ttk.Button(line, text='X', command=remove, width=1).pack(side=tk.LEFT, padx=5)
        update_val()

    def get_parameters(self) -> dict[str, str | float | tuple[float, float]]:
        pars = {}
        for var_name, var_lab, var_val, var_tol in self.vars:
            name = var_name.get()
            if name:
                value = var_val.get()
                tol = var_tol.get()
                try:
                    value = float(value)
                except ValueError:
                    pass
                try:
                    tol = float(tol)
                except ValueError:
                    tol = None
                if tol:
                    pars[name] = (value, tol)
                else:
                    pars[name] = value
        logger.debug(f"FindScans parameters: {pars}")
        return pars

    def find_scans(self):
        pars = self.get_parameters()
        scans = self.exp.find_scans(hdf_map=self.hdf_map, first_only=self.return_on_first, **pars)
        self.scan_numbers = [scan.scan_number() for scan in scans]
        logger.debug(f"found scan numbers: {self.scan_numbers}")
        self.close_fun()

    def wait_for_numbers(self) -> list[int]:
        self.root.wait_window()
        logger.debug('Returning scan numbers')
        return self.scan_numbers

    def wait_for_number(self) -> list[int]:
        self.return_on_first = True
        self.root.wait_window()
        logger.debug('Returning first scan number')
        return self.scan_numbers

    def wait_for_files(self) -> dict[int, str]:
        self.root.wait_window()
        logger.debug('Returning scan files')
        return {n: self.exp.scan_list[n] for n in self.scan_numbers}



"""
A python editor window
"""

import re

from ..misc.styles import tk, ttk
from ..misc.search import search_text
from ..misc.logging import create_logger

logger = create_logger(__file__)


# Define colors for the various types of tokens
class Colours:
    normal = '#eaeaea'  # rgb((234, 234, 234))
    keywords = '#ea5f5f'  # rgb((234, 95, 95))
    commands = '#54AAE3'
    comments = '#5feaa5'  # rgb((95, 234, 165))
    string = '#eaa25f'  # rgb((234, 162, 95))
    function = '#5fd3ea'  # rgb((95, 211, 234))
    background = '#2a2a2a'  # rgb((42, 42, 42))
    input = '#909184'  # rbb((0, 0, 255))
    highlight = '#baacab' # rgb 186, 172, 171

FONT = 'Consolas 15'

# Define a list of Regex Pattern that should be colored in a certain way
REPL = [
    [
        r'(?:^|\s|\W)(Scan|WARNING|ERROR)(?:$|\s|\W)',
        Colours.keywords
    ],
    [r'\A.*\| >>>.*\Z', Colours.input],
    # [r'(?:^|\s|\W)(scan|scancn|cscan|frange|pos|inc|go)(?:$|\s|\W)', commands],
    ['".*?"', Colours.string],
    ['\'.*?\'', Colours.string],
    ['#.*?$', Colours.comments],
]


def search_re(pattern, text):
    """Search text for regex pattern"""
    matches = []
    regex = re.compile(pattern)
    for i, line in enumerate(text.splitlines()):
        for match in regex.finditer(line):
            if match.groups():
                matches.append((f"{i + 1}.{match.span(1)[0]}", f"{i + 1}.{match.span(1)[1]}"))
            else:
                matches.append((f"{i + 1}.{match.start()}", f"{i + 1}.{match.end()}"))
    return matches


def update_tags(text: tk.Text):
    """ apply tags to textbox """

    # Remove all tags so they can be redrawn
    for tag in text.tag_names():
        text.tag_remove(tag, "1.0", tk.END)

    # Add tags where the search_re function found the pattern
    i = 0
    for pattern, color in REPL:
        for start, end in search_re(pattern, text.get('1.0', tk.END)):
            text.tag_add(f'{i}', start, end)
            text.tag_config(f'{i}', foreground=color)
            i += 1

    # Add tags to multiline comments
    start = None
    for n, line in enumerate(text.get('1.0', tk.END).splitlines()):
        for match in re.finditer('\'{3}|\"{3}', line):
            if start:
                text.tag_add(f'{i}', start, f"{n + 1}.{match.end()}")
                text.tag_config(f'{i}', foreground=Colours.comments)
                i += 1
                start = None
            else:
                start = f"{n + 1}.{match.start()}"


def log_tab(root: tk.Misc, log_string: str):
    """Create log textbox"""

    "----------- Textbox -----------"

    txt = ttk.Frame(root)
    txt.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

    # Scrollbars
    scroll_x = ttk.Scrollbar(txt, orient=tk.HORIZONTAL)
    scroll_x.pack(side=tk.BOTTOM, fill=tk.X)
    scroll_y = ttk.Scrollbar(txt, orient=tk.VERTICAL)
    scroll_y.pack(side=tk.RIGHT, fill=tk.Y)

    # TEXT box
    text = tk.Text(
        txt,
        background=Colours.background,
        foreground=Colours.normal,
        insertbackground=Colours.normal,
        exportselection=True,
        relief=tk.FLAT,
        font=FONT,
        wrap=tk.NONE
    )
    text.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
    text.insert('1.0', log_string)
    text.config(state=tk.DISABLED)

    # make scrollbars work
    text.config(xscrollcommand=scroll_x.set, yscrollcommand=scroll_y.set)
    scroll_x.config(command=text.xview)
    scroll_y.config(command=text.yview)

    update_tags(text)
    return text


class LogViewerWidget:
    """
    Editable textbox with numbers at side and key bindings for Python
    """

    def __init__(self, root: tk.Misc, log_tabs: dict[str, list[str]]):

        self.root = root
        self.search_box = tk.StringVar(self.root, '')
        self.search_matchcase = tk.BooleanVar(self.root, False)
        self.search_all_dates = tk.BooleanVar(self.root, True)
        self.search_number = tk.StringVar(self.root, '')

        main = ttk.Frame(root)
        main.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        self.ini_search(main)

        frm = ttk.Frame(main)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        # Tabs
        self.view_tabs = ttk.Notebook(frm)
        self.tab_texts = []
        for title, log in log_tabs.items():
            tab = ttk.Frame(self.view_tabs)
            self.view_tabs.add(tab, text=title)
            self.tab_texts.append(log_tab(tab, '\n'.join(log)))
        self.view_tabs.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

    def ini_search(self, frame: tk.Misc):
        frm = ttk.Frame(frame)
        frm.pack(side=tk.TOP, anchor=tk.E)

        var = ttk.Entry(frm, textvariable=self.search_box, width=40)
        var.pack(side=tk.LEFT)
        # var.bind('<KeyRelease>', self.fun_search)
        var.bind('<Return>', self.fun_search)
        var.bind('<KP_Enter>', self.fun_search)
        var = ttk.Button(frm, text='Search', command=self.fun_search, width=10)
        var.pack(side=tk.LEFT)

        ttk.Checkbutton(frm, variable=self.search_matchcase, text='Case').pack(side=tk.LEFT)
        ttk.Checkbutton(frm, variable=self.search_all_dates, text='All Dates').pack(side=tk.LEFT)
        ttk.Label(frm, textvariable=self.search_number).pack(side=tk.LEFT)


    def fun_search(self, event=None):
        """Search currently active tab"""
        found = 0
        if self.search_all_dates.get():
            # search all tabs
            first_tab = False
            for tab_index, tab_text in enumerate(self.tab_texts):
                found += search_text(tab_text, self.search_box.get(), self.search_matchcase.get(), Colours.highlight)
                if not first_tab and found > 0:
                    self.view_tabs.select(tab_index)
                    first_tab = True
        else:
            tab_index = self.view_tabs.index(self.view_tabs.select())
            tab_text = self.tab_texts[tab_index]
            found += search_text(tab_text, self.search_box.get(), self.search_matchcase.get(), Colours.highlight)
        self.search_number.set(f"{found} found")





"""

"""

import tkinter as tk

from mmg_toolbox.tkguis.misc.functions import topmenu, select_hdf_file
from mmg_toolbox.tkguis.misc.styles import RootWithStyle, create_root


def create_nexus_viewer(filename: str, parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    File Viewer of the NeXus structure
    """
    from mmg_toolbox.tkguis.widgets.hdf_viewer import HDFViewer

    root = create_root(parent=parent, window_title='NeXus File Viewer')

    widget = HDFViewer(root, filename, config)

    menu = {
        'File': {
            'Open': widget.select_file
        },
    }
    topmenu(root, menu, add_themes=False, add_about=True)

    if parent is None:
        root.mainloop()
    return root


def create_nexus_plotter(*filenames: str, parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    Plot 1D line data from a NeXus file
    """
    from ..widgets.nexus_plot import NexusDefaultPlot
    from mmg_toolbox.tkguis.apps.config_editor import ConfigEditor

    root = create_root('NeXus File Default Plot', parent=parent)
    widget = NexusDefaultPlot(root, *filenames, config=config)

    def load_file():
        new_filename = select_hdf_file(root)
        if new_filename:
            widget.update_data_from_files(new_filename)

    menu = {
        'File': {
            'Open': load_file,
        },
        'Config.': {
            'Edit Config.': lambda: ConfigEditor(root, config),
        }
    }

    topmenu(root, menu, add_themes=False, add_about=True)
    if parent is None:
        root.mainloop()
    return root


def create_nexus_image_plotter(filename: str, parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    Plot 2D images from detectors in a NeXus file, including a slider
    """
    from ..widgets.nexus_image import NexusDetectorImage

    root = create_root(parent=parent, window_title='NeXus File Image Plot')

    widget = NexusDetectorImage(root, filename, config)

    def load_file():
        new_filename = select_hdf_file(root)
        if new_filename:
            widget.update_image_data_from_file(new_filename)

    menu = {
        'File': {
            'Open': load_file,
        },
    }
    topmenu(root, menu, add_themes=False, add_about=True)

    if parent is None:
        root.mainloop()
    return root


def create_nexus_plot_and_image(*filenames: str, parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    Plot 1D line and image data from a NeXus file
    """
    from ..widgets.nexus_plot_and_image import NexusPlotAndImage
    from mmg_toolbox.tkguis.apps.config_editor import ConfigEditor

    root = create_root('NeXus File Default Plot', parent=parent)
    widget = NexusPlotAndImage(root, *filenames, config=config, horizontal_alignment=True)

    def load_file():
        new_filename = select_hdf_file(root)
        if new_filename:
            widget.update_data_from_files(new_filename)

    menu = {
        'File': {
            'Open': load_file,
        },
        'Config.': {
            'Edit Config.': lambda: ConfigEditor(root, config),
        }
    }

    topmenu(root, menu, add_themes=False, add_about=True)
    if parent is None:
        root.mainloop()
    return root





import os
import tkinter as tk

from mmg_toolbox.utils.env_functions import get_notebook_directory, get_scan_number
from ..misc.config import get_config, C
from ..misc.functions import topmenu
from ..misc.styles import RootWithStyle, create_root
from ..misc.processing_options import generate_replacement_getter, generate_processing_menu


def create_data_viewer(initial_folder: str | None = None,
                       parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    Create a Data Viewer showing all scans in an experiment folder
    """
    from ..widgets.nexus_data_viewer import NexusDataViewer
    from .log_viewer import create_gda_terminal_log_viewer
    from .file_browser import create_nexus_file_browser, create_file_browser, create_jupyter_browser
    from .scans import create_range_selector

    root = create_root(parent=parent, window_title='NeXus Data Viewer')
    config = config or get_config()

    widget = NexusDataViewer(root, initial_folder=initial_folder, config=config)

    def get_filepath():
        filename, folder = widget.selector_widget.get_filepath()
        return folder

    def get_scannos():
        return [get_scan_number(f) for f in widget.selector_widget.get_multi_filepath()]

    getter = generate_replacement_getter(
        scanno_getter=get_scannos,
        x_getter=widget.plot_widget.axes_x.get,
        y_getter=widget.plot_widget.axes_y.get,
        title_getter=lambda: f"Example Script: {os.path.basename(get_filepath())}"
    )
    processing_menu = generate_processing_menu(
        parent=root,
        config=config,
        directory=initial_folder,
        scan_files_getter=widget.selector_widget.get_multi_filepath,
        replacement_getter=getter
    )

    menu = {
        'File': {
            'New Data Viewer': lambda: create_data_viewer(parent=root, config=config),
            'Add Folder': widget.selector_widget.browse_folder,
            'File Browser': lambda: create_file_browser(root, config.get(C.default_directory, None)),
            'NeXus File Browser': lambda: create_nexus_file_browser(root, config.get(C.default_directory, None)),
            'Jupyter Browser': lambda: create_jupyter_browser(root, get_notebook_directory(get_filepath())),
            'Range selector': lambda: create_range_selector(initial_folder, root, config),
            'Log viewer': lambda: create_gda_terminal_log_viewer(get_filepath(), root)
        },
        'Processing': processing_menu
    }
    menu.update(widget.plot_widget.options_menu())

    topmenu(root, menu, add_themes=True, add_about=True, config=config)

    root.update()

    if parent is None:
        root.mainloop()
    return root



"""
mmg_toolbox tests
Test nx transformations
"""

from mmg_toolbox.nexus.instrument_model import NXInstrumentModel
from mmg_toolbox.nexus.nexus_reader import read_nexus_file
from . import only_dls_file_system
from .example_files import DIR

@only_dls_file_system
def test_read_nexus_file():
    f = DIR + 'i16/1116988.nxs'
    scan = read_nexus_file(f)
    model = scan.instrument_model()
    assert isinstance(model, NXInstrumentModel)




"""
tk widgets for viewing text based nexus data
"""

import tkinter as tk
from tkinter import ttk

import hdfmap

from mmg_toolbox.tkguis.misc.styles import update_text_style


class _StringView:
    """
    String viewer widget for NeXus file viewer
    """

    def __init__(self, root: tk.Misc):

        frm = ttk.Frame(root)
        frm.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

        self.text = tk.Text(frm, wrap=tk.NONE)

        vbar = ttk.Scrollbar(frm, orient=tk.VERTICAL, command=self.text.yview)
        hbar = ttk.Scrollbar(frm, orient=tk.HORIZONTAL, command=self.text.xview)
        self.text.configure(yscrollcommand=vbar.set, xscrollcommand=hbar.set)

        hbar.pack(side=tk.BOTTOM, fill=tk.X)
        vbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.text.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

        if hasattr(root, 'style'):
            update_text_style(self.text, root.style)

    def _populate(self, string):
        self.delete()
        self.text.insert('1.0', string)

    def populate(self, **kwargs):
        pass

    def delete(self):
        self.text.delete('1.0', tk.END)


class HdfTreeStr(_StringView):
    """
    HDF Tree String object
    """
    def populate(self, hdf_filename: str):
        """Load HDF file, populate ttk.treeview object"""
        tree_str = hdfmap.hdf_tree_string(hdf_filename)
        self._populate(tree_str)


class HdfNexusStr(_StringView):
    """
    Nexus Info String object
    """

    def populate(self, hdf_map: hdfmap.NexusMap):
        """Load HDF file, populate ttk.treeview object"""
        tree_str = hdf_map.info_nexus()
        self._populate(tree_str)


class Nexus2SrsStr(_StringView):
    """
    Display output of NeXus2SRS
    """

    def populate(self, hdf_map: hdfmap.NexusMap):
        """Load HDF file, populate ttk.treeview object"""
        from nexus2srs.nexus2srs import generate_datafile
        with hdf_map.load_hdf() as hdf:
            outstr, detector_image_paths = generate_datafile(hdf, hdf_map)
        self._populate(outstr)


class NxTransformationsStr(_StringView):
    """
    Display output of NeXus2Transformations
    """
    def populate(self, hdf_filename: str):
        from mmg_toolbox.nexus.nexus_transformations import generate_nxtranformations_string
        outstr = generate_nxtranformations_string(hdf_filename)
        self._populate(outstr)



"""
Polarisation utilities
"""

import numpy as np
import h5py

from mmg_toolbox.nexus.nexus_functions import nx_find, bytes2str
from mmg_toolbox.nexus.nexus_names import NX_POLARISATION_FIELDS


class PolLabels:
    linear_horizontal = 'lh'
    linear_vertical = 'lv'
    linear_arbitrary = 'la'
    circular_left = 'cl'
    circular_right = 'cr'
    circular_positive = 'pc'  # == circular_right
    circular_negative = 'nc'  # == circular_left
    linear_dichroism = 'xmld'
    circular_dichroism = 'xmcd'


def stokes_from_vector(*parameters: float) -> tuple[float, float, float, float]:
    """
    Return the Stokes parameters from an n-length vector
    """
    if len(parameters) == 4:
        p0, p1, p2, p3 = parameters
    elif len(parameters) == 3:
        p0 = 1
        p1, p2, p3 = parameters
    elif len(parameters) == 2:
        # polarisation vector [h, v]
        h, v = parameters
        p0, p3 = 1, 0
        phi = np.arctan2(v, h)
        p1, p2 = np.cos(2*phi), np.sin(2*phi)
    elif len(parameters) == 1:
        phi = np.deg2rad(parameters[0])
        p0, p3 = 1, 0
        p1, p2 = np.cos(2 * phi), np.sin(2 * phi)
    else:
        raise ValueError(f"Stokes parameters wrong length: {parameters}")
    return p0, p1, p2, p3


def polarisation_label_from_stokes(*stokes_parameters: float) -> str:
    """Convert Stokes vector to polarisation mode"""
    p0, p1, p2, p3 = stokes_from_vector(*stokes_parameters)
    circular = abs(p3) > 0.1
    if not circular and p1 > 0.9:
        return PolLabels.linear_horizontal
    if not circular and p1 < -0.9:
        return PolLabels.linear_vertical
    if not circular and np.sqrt(p1**2 + p2**2) > 0.9:
        return PolLabels.linear_arbitrary
    if circular and p3 > 0:
        return PolLabels.circular_right
    if circular and p3 < 0:
        return PolLabels.circular_left
    raise ValueError(f"Stokes parameters not recognized: {stokes_parameters}")


def polarisation_label_to_stokes(label: str, arbitrary_angle: float | None = None) -> tuple[float, float, float, float]:
    """Convert polarisation mode to Stokes vector"""
    label = bytes2str(label).strip().lower()
    match label:
        case PolLabels.linear_horizontal:
            return 1, 1, 0, 0
        case PolLabels.linear_vertical:
            return 1, -1, 0, 0
        case PolLabels.circular_right:
            return 1, 0, 0, 1
        case PolLabels.circular_left:
            return 1, 0, 0, -1
        case PolLabels.linear_arbitrary:
            if arbitrary_angle is not None:
                return stokes_from_vector(arbitrary_angle)
            raise ValueError("Linear arbitrary polarisation requires arbitrary_angle")
        # assume positive-circular is right-handed
        case PolLabels.circular_positive:
            return 1, 0, 0, 1
        case PolLabels.circular_negative:
            return 1, 0, 0, -1
    return 1, 0, 0, 0


def check_polarisation(label: str | np.ndarray | None, arbitrary_angle: float | None = None) -> str:
    """Return regularised polarisation mode"""
    if isinstance(label, str):
        return polarisation_label_from_stokes(*polarisation_label_to_stokes(label, arbitrary_angle))
    if isinstance(label, np.ndarray):
        return polarisation_label_from_stokes(*stokes_from_vector(*label))
    if label is None and arbitrary_angle is not None:
        return polarisation_label_from_stokes(*stokes_from_vector(arbitrary_angle))
    raise ValueError(f"Polarisation parameters not recognized: {label}")


def get_polarisation(pol: h5py.Dataset | h5py.Group) -> str:
    """
    Return polarisation mode from h5py Dataset, Group or File

    Raises ValueError if polarisation not recognized.

    Example:
        with h5py.File('data.nxs', 'r') as hdf:
            pol = get_polarisation(hdf)
            # -or-
            dataset = nx_find(hdf, 'NXbeam', 'incident_polarization_stokes')
            pol = get_polarisation(dataset)

    Parameters:
    :param pol: h5py.Dataset or h5py.Group object
    :return: polarisation mode
    """
    if isinstance(pol, h5py.Group):
        for label in NX_POLARISATION_FIELDS:
            dataset = nx_find(pol, label)
            if dataset:  # DLS specific polarisation mode
                return get_polarisation(dataset)
        raise KeyError(f"{pol} contains no polarisation fields")
    if np.issubdtype(pol.dtype, np.number):
        if pol.size == 1:
            return polarisation_label_from_stokes(pol[...])
        return polarisation_label_from_stokes(*pol)
    return check_polarisation(pol[()])


def pol_subtraction_label(label: str):
    """Return xmcd or xmld"""
    label = check_polarisation(label)
    if label in [PolLabels.linear_horizontal, PolLabels.linear_vertical]:
        return PolLabels.linear_dichroism
    elif label in [PolLabels.circular_left, PolLabels.circular_right]:
        return PolLabels.circular_dichroism
    else:
        raise ValueError(f"Polarisation label not recognized: {label}")


def analyser_jones_matrix(crystal_bragg: float, rotation: float) -> np.ndarray:
    """
    Return the Jones matrix for an analyser crystal
    The Jones matrix [4x4] can be used to operate on a Jones vector [2x1]
    to describe how the polarisation will be analysed.

    The basis chosen in such that x = v X z, where v is vertical and z is the incident beam direction in lab space.

    :param crystal_bragg: Scattering Bragg angle, in degrees
    :param rotation: Rotation angle about the incident beam, in degrees
    :return: Jones matrix [4x4]
    """
    bragg_rad = np.deg2rad(crystal_bragg)
    rot_rad = np.deg2rad(rotation)
    return np.array([
        [np.cos(rot_rad), -np.sin(rot_rad)],
        [np.sin(rot_rad) * np.cos(bragg_rad), np.cos(rot_rad) * np.cos(bragg_rad)]
    ])


#TODO: Check this!
def jones2mueller(jones: np.ndarray) -> np.ndarray:
    """
    Convert Jones matrix [2x2] to Mueller matrix [4x4]
    see https://en.wikipedia.org/wiki/Mueller_calculus
    """
    A = np.array([
        [1, 0, 0, 1],
        [1, 0, 0, -1],
        [0, 1, 1, 0],
        [0, 1j, -1j, 0],
    ], dtype=complex)
    mueller = np.linalg.multi_dot([A, np.kron(jones, np.conj(jones)), np.linalg.inv(A)])
    return mueller


#TODO: Check this!
def apply_jones_to_stokes(S: np.ndarray, J: np.ndarray) -> np.ndarray:
    """
    Applies a Jones matrix to a Stokes vector and returns the resulting Stokes vector.

    Parameters:
    S (array-like): Input Stokes vector [I, Q, U, V]
    J (2x2 array): Jones matrix

    Returns:
    numpy.ndarray: Transformed Stokes vector [I', Q', U', V']
    """
    I, Q, U, V = S

    # Convert Stokes to Jones vector (assuming fully polarized light)
    Ex = np.sqrt((I + Q) / 2)
    Ey = np.sqrt((I - Q) / 2)
    phase = np.arctan2(V, U)
    Ey *= np.exp(1j * phase)
    E = np.array([Ex, Ey])

    # Apply Jones matrix
    E_out = J @ E

    # Convert back to Stokes parameters
    I_out = np.real(E_out[0]*np.conj(E_out[0]) + E_out[1]*np.conj(E_out[1]))
    Q_out = np.real(E_out[0]*np.conj(E_out[0]) - E_out[1]*np.conj(E_out[1]))
    U_out = 2 * np.real(E_out[0]*np.conj(E_out[1]))
    V_out = 2 * np.imag(E_out[0]*np.conj(E_out[1]))

    return np.array([I_out, Q_out, U_out, V_out])


#TODO: Check this!
def analyse_polarisation(stokes_vector: np.ndarray, *jones_matrices: np.ndarray) -> np.ndarray:
    """
    Applies a Jones matrix to a Stokes vector and returns the resulting Stokes vector.

    Parameters:
    stokes_vector (1x4 array): incident beam Stokes parameters P0, P1, P2, P3 (I, Q, U, V)
    jones_matrix (2x2 array): Jones matrix of the optical element
    * multiple jones_matrices can be entered and will be applied in order

    Returns:
    numpy.ndarray: Transformed Stokes vector [I', Q', U', V']
    """
    I, Q, U, V = stokes_vector

    # Convert Stokes to Jones vector (assuming fully polarized light)
    Ex = np.sqrt((I + Q) / 2)
    Ey = np.sqrt((I - Q) / 2)
    phase = np.arctan2(V, U)
    Ey *= np.exp(1j * phase)
    E_in = np.array([Ex, Ey])

    # Apply Jones matrix
    # E_out = jones_matrices[0] @ E_in
    E_out = np.linalg.multi_dot(list(jones_matrices) + [E_in])

    print(f"Ein={E_in} -> Eout={E_out}")

    # Convert back to Stokes parameters
    I_out = np.real(E_out[0]*np.conj(E_out[0]) + E_out[1]*np.conj(E_out[1]))
    Q_out = np.real(E_out[0]*np.conj(E_out[0]) - E_out[1]*np.conj(E_out[1]))
    U_out = 2 * np.real(E_out[0]*np.conj(E_out[1]))
    V_out = 2 * np.imag(E_out[0]*np.conj(E_out[1]))

    return np.array([I_out, Q_out, U_out, V_out])


#TODO: Check this!
def stokes_to_lab_vector(S, beam_direction, reference_plane):
    """
    ***CHECK THIS!!!***
    Converts Stokes parameters to a 3D polarization vector in lab space.

    Parameters:
    S (array-like): Stokes vector [I, Q, U, V]
    beam_direction (array-like): 3D unit vector of beam propagation direction
    reference_plane (array-like): 3D vector defining the reference plane (e.g., horizontal)

    Returns:
    numpy.ndarray: 3D polarization vector in lab space
    """
    # Normalize input vectors
    k = np.array(beam_direction, dtype=float)
    k /= np.linalg.norm(k)
    ref = np.array(reference_plane, dtype=float)
    ref -= np.dot(ref, k) * k  # Make ref orthogonal to k
    ref /= np.linalg.norm(ref)

    # Define orthonormal basis for transverse plane
    s = ref  # s-polarization (in reference plane)
    p = np.cross(k, s)  # p-polarization (perpendicular to s and k)

    # Convert Stokes to Jones vector (assuming fully polarized light)
    I, Q, U, V = S
    Ex = np.sqrt((I + Q) / 2)
    Ey = np.sqrt((I - Q) / 2)
    phase = np.arctan2(V, U)
    Ey *= np.exp(1j * phase)
    E = np.array([Ex, Ey])

    # Construct 3D polarization vector in lab frame
    E_lab = E[0] * s + E[1] * p

    return E_lab




"""
Peak functions
"""

import numpy as np


__all__ = ['poisson_errors', 'peak_ratio', 'gen_weights', 'gauss', 'group_adjacent', 'find_peaks', 'find_peaks_str', 'max_index']


def poisson_errors(y: np.ndarray) -> np.ndarray:
    """Default error function for counting statistics"""
    return np.sqrt(np.abs(y) + 1)


def peak_ratio(y: np.ndarray, yerror: np.ndarray | None = None) -> float:
    """
    Return the ratio signal / error for given dataset
    From Blessing, J. Appl. Cryst. (1997). 30, 421-426 Equ: (1) + (6)
      peak_ratio = (sum((y-bkg)/dy^2)/sum(1/dy^2)) / sqrt(i/sum(1/dy^2))
    :param y: array of y data
    :param yerror: array of errors on data, or None to calcualte np.sqrt(y+0.001)
    :return: float ratio signal / err
    """
    if yerror is None:
        yerror = poisson_errors(y)
    bkg = np.min(y)
    wi = 1 / yerror ** 2
    signal = np.sum(wi * (y - bkg)) / np.sum(wi)
    err = np.sqrt(len(y) / np.sum(wi))
    return signal / err


def gen_weights(yerrors=None) -> np.ndarray | None:
    """
    Generate weights for fitting routines
    :param yerrors: array(n) or None
    :return: array(n) or None
    """
    if yerrors is None or np.all(np.abs(yerrors) < 0.001):
        weights = None
    else:
        yerrors = np.asarray(yerrors, dtype=float)
        yerrors[yerrors < 1] = 1.0
        weights = 1 / yerrors
        weights = np.abs(np.nan_to_num(weights))
    return weights


def gauss(x: np.ndarray, y: np.ndarray | None = None,
          height: float = 1, cen: float = 0, fwhm: float = 0.5,
          bkg: float = 0, cen_y: float | None = None,
          fwhm_y: float | None = None) -> np.ndarray:
    """
    Define Gaussian distribution in 1 or 2 dimensions

        y[1xn] = gauss(x[1xn], height=10, cen=0, fwhm=1, bkg=0)
           - OR -
        Z[nxm] = gauss(x[1xn], y[1xm], height=100, cen=4, fwhm=5, bkg=30)

    From http://fityk.nieto.pl/model.html

    height is related to amplitude (area) by:
        height = (area / fwhm) * sqrt(4*ln(2)/pi) ~ 0.94 * area / fwhm
        area = height * fwhm /  sqrt(4*ln(2)/pi) ~ 1.06 * height * fwhm
        - or for 2D -
        area = height * fwhm_x * fwhm_y * (pi / 4*ln(2))
    sigma is related to fwhm by:
        fwhm = 2 * sqrt(2*ln(2)) * sigma
        sigma = fwhm / (2*sqrt(2*ln(2)))

    :param x: [1xn] array of values, defines size of gaussian in dimension 1
    :param y: None* or [1xm] array of values, defines size of gaussian in dimension 2
    :param height: peak height
    :param cen: peak centre
    :param fwhm: peak full width at half-max
    :param bkg: background
    :param cen_y: peak centre in y-axis (None to use cen)
    :param fwhm_y: peak full width in y-axis (None to use fwhm)
    :returns: [1xn array] Gassian distribution
    - or, if y is not None: -
    :returns: [nxm array] 2D Gaussian distribution
    """

    cen_x = cen
    fwhm_x = fwhm
    if cen_y is None:
        cen_y = cen
    if fwhm_y is None:
        fwhm_y = fwhm
    if y is None:
        y = cen_y

    x = np.asarray(x, dtype=float).reshape([-1])
    y = np.asarray(y, dtype=float).reshape([-1])
    X, Y = np.meshgrid(x, y)
    g = height * np.exp(
        -np.log(2) * (
                ((X - cen_x) ** 2) / (fwhm_x / 2) ** 2 +
                ((Y - cen_y) ** 2) / (fwhm_y / 2) ** 2
        )
    ) + bkg

    if len(y) == 1:
        g = g.reshape([-1])
    return g


def group_adjacent(values: np.ndarray, close: float = 10):
    """
    Average adjacent values in array, return grouped array and indexes to return groups to original array

    E.G.
     grp, idx = group_adjacent([1,2,3,10,12,31], close=3)
     grp -> [2, 11, 31]
     idx -> [[0,1,2], [3,4], [5]]

    :param values: array of values to be grouped
    :param close: float
    :return grouped_values: float array(n) of grouped values
    :return indexes: [n] list of lists, each item relates to an averaged group, with indexes from values
    """
    # Check distance between good peaks
    dist_chk = []
    dist_idx = []
    gx = 0
    dist = [values[gx]]
    idx = [gx]
    while gx < len(values) - 1:
        gx += 1
        if (values[gx] - values[gx - 1]) < close:
            dist += [values[gx]]
            idx += [gx]
        else:
            dist_chk += [np.mean(dist)]
            dist_idx += [idx]
            dist = [values[gx]]
            idx = [gx]
    dist_chk += [np.mean(dist)]
    dist_idx += [idx]
    return np.array(dist_chk), dist_idx


def local_maxima_1d(y: np.ndarray) -> np.ndarray:
    """
    Find local maxima in 1d array
    Returns points with central point higher than neighboring points.

    Copied from scipy.signal._peak_finding_utils
    https://github.com/scipy/scipy/blob/v1.7.1/scipy/signal/_peak_finding_utils.pyx

    :param y: list or array
    :return: array of peak indexes
    """
    y = np.asarray(y, dtype=float).reshape(-1)

    # Preallocate, there can't be more maxima than half the size of `y`
    midpoints = np.empty(y.shape[0] // 2, dtype=np.intp)
    m = 0  # Pointer to the end of valid area in allocated arrays
    i = 1  # Pointer to current sample, first one can't be maxima
    i_max = y.shape[0] - 1  # Last sample can't be maxima
    while i < i_max:
        # Test if previous sample is smaller
        if y[i - 1] < y[i]:
            i_ahead = i + 1  # Index to look ahead of current sample

            # Find next sample that is unequal to x[i]
            while i_ahead < i_max and y[i_ahead] == y[i]:
                i_ahead += 1

            # Maxima is found if next unequal sample is smaller than x[i]
            if y[i_ahead] < y[i]:
                left_edge = i
                right_edge = i_ahead - 1
                midpoints[m] = (left_edge + right_edge) // 2
                m += 1
                # Skip samples that can't be maximum
                i = i_ahead
        i += 1
    return midpoints[:m]


def find_local_maxima(y: np.ndarray, yerror: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Find local maxima in 1d arrays, returns index of local maximums, plus
    estimation of the peak power for each maxima and a classification of whether the maxima is greater than
    the standard deviation of the error.

    E.G.
        index, power, isgood = find_local_maxima(ydata)
        maxima = ydata[index[isgood]]
        maxima_power = power[isgood]

    Peak Power:
      peak power for each maxima is calculated using the peak_ratio algorithm for each maxima and adjacent points
    Good Peaks:
      Maxima are returned Good if:  power > (max(y) - min(y)) / std(yerror)

    :param y: array(n) of data
    :param yerror: array(n) of errors on data, or None to use default error function (sqrt(abs(y)+1))
    :return index: array(m<n) of indexes in y of maxima
    :return power: array(m) of estimated peak power for each maxima
    :return isgood: bool array(m) where True elements have power > power of the array
    """

    if yerror is None or np.all(np.abs(yerror) < 0.1):
        yerror = poisson_errors(y)
    else:
        yerror = np.asarray(yerror, dtype=float)
    yerror[yerror < 1] = 1.0
    bkg = np.min(y)
    wi = 1 / yerror ** 2

    index = local_maxima_1d(y)
    # average nearest 3 points to peak
    power = np.array([np.sum(wi[m - 1:m + 2] * (y[m - 1:m + 2] - bkg)) / np.sum(wi[m - 1:m + 2]) for m in index])
    # Determine if peak is good
    isgood = power > (np.max(y) - np.min(y)) / (np.std(yerror) + 1)
    return index, power, isgood


def find_peaks(y: np.ndarray, yerror: np.ndarray | None = None,
               min_peak_power: float | None = None, peak_distance_idx: int = 6) -> tuple[np.ndarray, np.ndarray]:
    """
    Find peak shaps in linear-spaced 1d arrays with poisson like numerical values

    E.G.
      index, power = find_peaks(ydata, yerror, min_peak_power=None, peak_distance_idx=10)
      peak_centres = xdata[index]  # ordered by peak strength

    :param y: array(n) of data
    :param yerror: array(n) of errors on data, or None to use default error function (sqrt(abs(y)+1))
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :return index: array(m) of indexes in y of peaks that satisfy conditions
    :return power: array(m) of estimated power of each peak
    """
    # Get all peak positions
    midpoints, peak_signals, chk = find_local_maxima(y, yerror)

    if min_peak_power is None:
        good_peaks = chk
    else:
        good_peaks = peak_signals >= min_peak_power

    # select indexes of good peaks
    peaks_idx = midpoints[good_peaks]
    peak_power = peak_signals[good_peaks]
    if len(peaks_idx) == 0:
        return peaks_idx, peak_power

    # Average peaks close to each other
    group_idx, group_signal_idx = group_adjacent(peaks_idx, peak_distance_idx)
    peaks_idx = np.round(group_idx).astype(int)
    peak_power = np.array([np.sum(peak_power[ii]) for ii in group_signal_idx])

    # sort peak order by strength
    power_sort = np.argsort(peak_power)[::-1]
    return peaks_idx[power_sort], peak_power[power_sort]


def find_peaks_str(x: np.ndarray, y: np.ndarray, yerror: np.ndarray | None = None,
                   min_peak_power: float | None = None, peak_distance_idx: int = 6) -> str:
    """
    Find peak shaps in linear-spaced 1d arrays with poisson like numerical values

    E.G.
      index, power = find_peaks(ydata, yerror, min_peak_power=None, peak_distance_idx=10)
      peak_centres = xdata[index]  # ordered by peak strength

    :param x: array(n) of data
    :param y: array(n) of data
    :param yerror: array(n) of errors on data, or None to use default error function (sqrt(abs(y)+1))
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :return index: array(m) of indexes in y of peaks that satisfy conditions
    :return power: array(m) of estimated power of each peak
    """
    index, power = find_peaks(y, yerror, min_peak_power, peak_distance_idx)
    x_vals = x[index]
    out = f"Find Peaks:\n len: {len(x)}, max: {np.max(y):.5g}, min: {np.min(y):.5g}\n\n"
    out += '\n'.join(f"  {idx:4} {_x:10.5}  power={pwr:.3}" for idx, _x, pwr in zip(index, x_vals, power))
    return out


def max_index(array: np.ndarray) -> tuple[int, ...]:
    """Return the index of the largest value in an array."""
    max_idx = np.nanargmax(array)
    return np.unravel_index(max_idx, array.shape)




"""
a tkinter frame with a single plot
"""
import os
import tkinter as tk
from tkinter import ttk
import numpy as np

import hdfmap
from hdfmap import create_nexus_map
from hdfmap.eval_functions import generate_identifier

from mmg_toolbox.utils.env_functions import get_scan_number
from mmg_toolbox.fitting import multipeakfit, FitResults, find_peaks_str
from ..misc.logging import create_logger
from ..misc.config import get_config
from .simple_plot import SimplePlot


logger = create_logger(__file__)


class NexusDefaultPlot(SimplePlot):
    """
    Tkinter widget for Nexus plot
    Widget contains 2D plot with single choice of x and y axes.

      widget  = NexusDefaultPlot(root, 'file.nxs', config)

    Axes can be chosen from a dropdown menu of the default scannables,
    or an entry box will be evaluated, allowing expressions.
    """
    def __init__(self, root: tk.Misc, *hdf_filenames: str, config: dict | None = None):
        self.root = root
        self.filenames = hdf_filenames
        self.map: hdfmap.NexusMap | None = None
        self.config = config or get_config()
        self._plot_data: list[dict] = []
        self._scannable_data: list[dict[str, np.ndarray]] = []  # plot data: list of dicts of arrays
        self._fit_result: FitResults | None = None

        self.axes_x = tk.StringVar(self.root, 'axes')
        self.axes_y = tk.StringVar(self.root, 'signal')
        self.normalise = tk.BooleanVar(self.root, False)
        self.fix_x = tk.BooleanVar(self.root, False)
        self.fix_y = tk.BooleanVar(self.root, False)
        self.fit_model = tk.StringVar(self.root, 'Gaussian')
        self.max_peaks = tk.IntVar(self.root, 1)
        self.do_fit = tk.BooleanVar(self.root, False)
        self.error_message = tk.StringVar(self.root, '')

        self.combo_x, self.combo_y = self.ini_axes_select()

        super().__init__(
            root=root,
            xdata=[],
            ydata=[],
            xlabel=self.axes_x.get(),
            ylabel=self.axes_y.get(),
            title='',
            config=config
        )
        self.line = self.plot_list[0]
        if hdf_filenames:
            self.update_data_from_files(*hdf_filenames)

    def _clear_error(self):
        self.error_message.set('')

    def _set_error(self, msg: str):
        self.error_message.set(msg)

    def plot(self, *args, **kwargs):
        lines = self.ax1.plot(*args, **kwargs)
        self.plot_list.extend(lines)

    def remove_lines(self):
        for obj in self.plot_list:
            obj.remove()
        self.plot_list.clear()

    def ini_axes_select(self):
        selection_x = tk.StringVar(self.root, 'axes')
        selection_y = tk.StringVar(self.root, 'signal')
        axes_options = ['axes', 'signal']
        signal_options = axes_options[::-1]

        def select_x(event):
            self.axes_x.set(selection_x.get())
            self.update_axis_choice()

        def select_y(event):
            self.axes_y.set(selection_y.get())
            self.update_axis_choice()

        section = ttk.Frame(self.root)
        section.pack(side=tk.TOP, expand=tk.NO, fill=tk.BOTH)

        frm = ttk.Frame(section)
        frm.pack(side=tk.LEFT)
        line = ttk.Frame(frm)
        line.pack(side=tk.TOP, expand=tk.NO, fill=tk.X)
        var = ttk.Label(line, text='X Axes:', width=10)
        var.pack(side=tk.LEFT)
        combo_x = ttk.Combobox(line, values=axes_options,
                               textvariable=selection_x, width=20)
        # combo_x.pack(side=tk.LEFT, padx=5)
        # combo_x.bind('<<ComboboxSelected>>', select_x)
        var = ttk.Entry(line, textvariable=self.axes_x, width=30)
        var.pack(side=tk.LEFT)
        # var.bind('<KeyRelease>', self.fun_expression_reset)
        var.bind('<Return>', self.update_axis_choice)
        var.bind('<KP_Enter>', self.update_axis_choice)
        var = ttk.Checkbutton(line, text='Fix', variable=self.fix_x)
        var.pack(side=tk.LEFT)

        line = ttk.Frame(frm)
        line.pack(side=tk.TOP, expand=tk.NO, fill=tk.X)
        var = ttk.Label(line, text='Y Axes:', width=10)
        var.pack(side=tk.LEFT)
        combo_y = ttk.Combobox(line, values=signal_options,
                               textvariable=selection_y, width=20)
        # combo_y.pack(side=tk.LEFT, padx=5)
        # combo_y.bind('<<ComboboxSelected>>', select_y)
        var = ttk.Entry(line, textvariable=self.axes_y, width=30)
        var.pack(side=tk.LEFT)
        # var.bind('<KeyRelease>', self.fun_expression_reset)
        var.bind('<Return>', self.update_axis_choice)
        var.bind('<KP_Enter>', self.update_axis_choice)
        var = ttk.Checkbutton(line, text='Fix', variable=self.fix_y)
        var.pack(side=tk.LEFT)
        var = ttk.Checkbutton(line, text='Normalise', variable=self.normalise, command=self.normalise_signal)
        var.pack(side=tk.LEFT)

        # Fitting
        frm = ttk.Frame(section, relief=tk.RIDGE, borderwidth=2)
        frm.pack(side=tk.LEFT, padx=4)
        line = ttk.Frame(frm)
        line.pack(side=tk.TOP, fill=tk.X)
        ttk.Label(line, text='Max peaks:').pack(side=tk.LEFT)
        var = ttk.Entry(line, textvariable=self.max_peaks, width=2)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', self.perform_fit)
        var.bind('<KP_Enter>', self.perform_fit)
        ttk.Checkbutton(line, variable=self.do_fit).pack(side=tk.LEFT)

        line = ttk.Frame(frm)
        line.pack(side=tk.TOP, fill=tk.X)
        fit_options = ['Line', 'Gaussian', 'Lorentzian', 'pVoight']  # TODO: get this from somewhere else
        var = ttk.Combobox(line, values=fit_options,
                           textvariable=self.fit_model, width=12)
        var.pack(side=tk.LEFT)
        var.bind('<<ComboboxSelected>>', self.perform_fit)
        ttk.Button(line, text=':', command=self.fit_results, width=1, padding=0).pack(side=tk.LEFT, padx=1)

        # Fitting
        frm = ttk.Frame(section)
        frm.pack(side=tk.LEFT, fill=tk.Y, padx=4)
        ttk.Button(frm, text='Processing', command=self.multiplots).pack(side=tk.TOP, fill=tk.X)
        ttk.Button(frm, text='Fitting', command=self.peakfiting).pack(side=tk.TOP, fill=tk.X)

        # Error line
        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, expand=tk.NO, fill=tk.X)
        ttk.Label(frm, textvariable=self.error_message, style='error.TLabel').pack()
        return combo_x, combo_y

    def update_data_from_files(self, *filenames: str, hdf_map: hdfmap.NexusMap | None = None):
        if not filenames:
            return
        self._clear_error()
        self.filenames = filenames
        self.map = create_nexus_map(filenames[0]) if hdf_map is None else hdf_map
        self._load_data()
        if not self._scannable_data:
            self._set_error("No data loaded")
            return
        first_dataset = self._scannable_data[0]
        self.combo_x['values'] = list(first_dataset.keys())
        self.combo_y['values'] = list(reversed(first_dataset.keys()))
        # Default axes choice
        axes, signals = self.map.nexus_default_names()
        if not self.fix_x.get():
            self.axes_x.set(next(iter(axes), f'arange({self.map.scannables_length()})'))
        if not self.fix_y.get():
            self.axes_y.set(next(iter(signals), f'zeros({self.map.scannables_length()})'))
        self.update_axis_choice()

    def _label(self, name: str) -> str:
        path = self.map.combined.get(name, '')
        if not path:
            return name
        units = self.map.get_attr(path, 'units', '')
        unit_str = f" [{units}]" if units else ''
        return generate_identifier(path) + unit_str

    def _load_data(self):
        self._plot_data = []
        self._scannable_data = []
        errors = []
        for filename in self.filenames:
            try:
                with hdfmap.load_hdf(filename) as hdf:
                    plot_data = self.map.get_plot_data(hdf)
            except Exception as e:
                errors.append(f"Error loading data in file {os.path.basename(filename)}: {e}")
            self._plot_data.append(plot_data)
            self._scannable_data.append(plot_data.get('data', {}))
        if errors:
            self._set_error('\n'.join(errors))

    def get_xy_data(self, x_label: str, *y_labels: str) -> tuple[list[np.ndarray], list[np.ndarray], list[str]]:
        x_data: list[np.ndarray] = []
        y_data: list[np.ndarray] = []
        labels: list[str] = []
        errors = []
        for filename, scannables in zip(self.filenames, self._scannable_data):
            this_x_data = scannables.get(x_label, None)
            this_y_data = [scannables.get(label, None) for label in y_labels]
            # TODO: handle 2D scans as additional set of labels
            file_label = f"#{get_scan_number(filename)}" if len(self.filenames) > 1 else ""
            this_labels = [f"{file_label} {lab}" for lab in y_labels] if len(y_labels) > 1 else [file_label]

            if this_x_data is None or any(data is None for data in this_y_data):
                # Load additional data
                with hdfmap.load_hdf(filename) as hdf:
                    if this_x_data is None:
                        try:
                            this_x_data = self.map.eval(hdf, x_label, np.arange(self.map.scannables_length()))
                        except Exception as e:
                            errors.append(f"Error loading x-axis in file {os.path.basename(filename)}: {e}")
                            this_x_data = np.arange(self.map.scannables_length())
                    for n, label in enumerate(y_labels):
                        if this_y_data[n] is None:
                            try:
                                this_y_data[n] = self.map.eval(hdf, label, np.arange(self.map.scannables_length()))
                            except Exception as e:
                                errors.append(f"Error loading y-axis in file {os.path.basename(filename)}: {e}")
                                this_y_data[n] = np.arange(len(this_x_data))
            x_data.extend([this_x_data] * len(this_y_data))
            y_data.extend(this_y_data)
            labels.extend(this_labels)

        if errors:
            self._set_error('\n'.join(errors))

        return x_data, y_data, labels

    def normalise_signal(self, event=None):
        signal = self.axes_y.get()
        norm_by = self.config.get('normalise_factor', '')
        if signal.endswith(norm_by):
            signal = signal.replace(norm_by, '')
        if self.normalise.get():
            self.axes_y.set(signal + norm_by)
        else:
            self.axes_y.set(signal)
        self.update_axis_choice()

    def update_axis_choice(self, event=None):
        x_label = self.axes_x.get()
        y_label = self.axes_y.get()
        if not x_label or not y_label:
            return
        xdata, ydata, labels = self.get_xy_data(x_label, *y_label.split(','))
        self.update_from_data(
            x_data=xdata,
            y_data=ydata,
            x_label=self._label(x_label),
            y_label=self._label(y_label),
            title=os.path.basename(self.filenames[0]),
            legend=labels,
        )
        self.line = self.plot_list[0]

    def _perform_fit(self) -> tuple[FitResults | None, str]:
        """Returns (FitResults, label)"""
        x_label = self.axes_x.get()
        y_label = self.axes_y.get()
        model = self.fit_model.get()
        peaks = self.max_peaks.get()
        if not x_label or not y_label:
            return None, ''
        xdata, ydata, labels = self.get_xy_data(x_label, y_label)
        result = multipeakfit(
            xvals=xdata[0],
            yvals=ydata[0],
            npeaks=peaks,
            model=model,
        )
        x_fit, y_fit = result.fit_data(ntimes=1)  # don't interpolate as x will be the wrong
        label = f"fit_{y_label}_{peaks}{model}"
        self._scannable_data[0][label] = y_fit
        self._fit_result = result
        return result, label

    def perform_fit(self, event=None):
        result, label = self._perform_fit()
        if result is None:
            return
        x, y = result.fit_data()
        lines = self.ax1.plot(x, y, label=label)
        self.plot_list.extend(lines)
        self.update_labels(legend=True)
        self.update_axes()

    def fit_options(self):
        pass

    def fit_results(self):
        from ..apps.edit_text import EditText

        x_axis = self.axes_x.get()
        y_axis = self.axes_y.get()
        if not x_axis or not y_axis:
            return
        xdata, ydata, labels = self.get_xy_data(x_axis, y_axis)
        peak_str = find_peaks_str(xdata[0], ydata[0])

        title = os.path.basename(self.filenames[0])
        x_label, y_label = self.map.generate_ids(x_axis, y_axis)
        label = f"{x_label} vs {y_label}"
        out = f"{title}\n{label}\n\n"
        out += peak_str + '\n\n'
        if self._fit_result is not None:
            out += str(self._fit_result)
        EditText(out, parent=self.root, title=title)

    def multiplots(self):
        from ..apps.multi_scan_analysis import create_multi_scan_analysis
        # Note that exp directory and proc directory are in config
        create_multi_scan_analysis(
            parent=self.root,
            config=self.config,
            scan_numbers=[get_scan_number(f) for f in self.filenames],
            x_axis=self.axes_x.get(),
            y_axis=self.axes_y.get(),
        )

    def peakfiting(self):
        from ..apps.peak_fit_analysis import create_peak_fit
        # Note that exp directory and proc directory are in config
        create_peak_fit(
            parent=self.root,
            config=self.config,
            scan_numbers=[get_scan_number(f) for f in self.filenames],
            x_axis=self.axes_x.get(),
            y_axis=self.axes_y.get(),
        )


class NexusMultiAxisPlot(NexusDefaultPlot):
    """
    1D line plot widget with selection box for selecting multiple axes.
    """
    def __init__(self, root: tk.Misc, *hdf_filenames: str,
                 config: dict | None = None):
        super().__init__(root, *hdf_filenames, config=config)
        self.fig.subplots_adjust(right=0.97)
        self._pause_selection = False
        self.listbox = self._axis_listbox()

    def _axis_listbox(self) -> ttk.Treeview:
        frame = ttk.Frame(self.root, width=150)
        frame.pack(side=tk.RIGHT, fill=tk.Y)
        frame.pack_propagate(False)

        y_scrollbar = ttk.Scrollbar(frame)
        x_scrollbar = ttk.Scrollbar(frame, orient="horizontal")
        listbox = ttk.Treeview(frame, yscrollcommand=y_scrollbar.set,
                               xscrollcommand=x_scrollbar.set, show="tree")
        listbox.column("#0", width=200, stretch=tk.YES)
        listbox.bind("<<TreeviewSelect>>", self.select_listbox_items)
        y_scrollbar.configure(command=listbox.yview)
        x_scrollbar.configure(command=listbox.xview)

        y_scrollbar.pack(side="right", fill="y")
        x_scrollbar.pack(side="bottom", fill="x")
        listbox.pack(side="left", fill="both", expand=True)
        return listbox

    def update_data_from_files(self, *filenames: str, hdf_map: hdfmap.NexusMap | None = None):
        super().update_data_from_files(*filenames, hdf_map=hdf_map)
        auto_signal = self.axes_y.get()

        # populate listbox
        self.listbox.delete(*self.listbox.get_children())
        first_dataset = self._scannable_data[0]
        # Temporarily unbind TreeviewSelect
        self._pause_selection = True
        for item in first_dataset:
            iid = self.listbox.insert("", tk.END, text=item)
            if item == auto_signal:
                self.listbox.selection_add(iid)
                self.listbox.focus(iid)
        self._pause_selection = False

    def select_listbox_items(self, event=None):
        if self._pause_selection or len(self.listbox.selection()) == 0:
            return
        self.remove_lines()
        x_label = self.axes_x.get()
        y_labels = [self.listbox.item(item)['text'] for item in self.listbox.selection()]
        self.axes_y.set(y_labels[0])
        xdata, ydata, labels = self.get_xy_data(x_label, *y_labels)
        self.update_from_data(
            x_data=xdata,
            y_data=ydata,
            x_label=self._label(x_label),
            y_label=self._label(y_labels[0]),
            title=os.path.basename(self.filenames[0]),
            legend=labels,
        )
        self.line = self.plot_list[0]
        if self.do_fit.get():
            self.perform_fit()

    def update_axis_choice(self, event=None):
        # select item in list if it matches
        yaxis = self.axes_y.get()
        iid = next((
            item for item in self.listbox.get_children()
            if yaxis == self.listbox.item(item)['text']
        ), None)
        if not iid:
            iid = self.listbox.insert("", tk.END, text=yaxis)
        self._pause_selection = True  # don't trigger TreeViewSelect
        self.listbox.selection_set(iid)
        self.listbox.see(iid)
        self._pause_selection = False
        super().update_axis_choice(event)

    def perform_fit(self, event=None):
        result, label = self._perform_fit()
        if result is None:
            return
        if label not in (self.listbox.item(iid, 'text') for iid in self.listbox.get_children()):
            iid = self.listbox.insert("", tk.END, text=label)
        else:
            iid = next(iid for iid in self.listbox.get_children() if self.listbox.item(iid, 'text') == label)
        self.listbox.selection_add(iid)




from mmg_toolbox.tkguis.misc.config import get_config, C
from mmg_toolbox.tkguis.misc.functions import topmenu
from mmg_toolbox.tkguis.misc.styles import create_root
from mmg_toolbox.tkguis.misc.jupyter import launch_jupyter_notebook, terminate_notebooks
from mmg_toolbox.utils.env_functions import open_terminal


def create_title_window(beamline: str | None = None):
    """Title Window"""
    from ..widgets.title_window import TitleWindow
    from .log_viewer import create_gda_terminal_log_viewer
    from .file_browser import create_nexus_file_browser, create_file_browser, create_jupyter_browser
    from .scans import create_range_selector
    from .data_viewer import create_data_viewer
    from .python_editor import create_python_editor
    from .visit_viewer import create_visit_viewer


    root = create_root(window_title='Beamline Data Viewer')
    config = get_config(beamline=beamline)

    widget = TitleWindow(root, config)

    menu = {
        'File': {
            'File Browser': lambda: create_file_browser(root, config.get(C.default_directory, None)),
            'NeXus File Browser': lambda: create_nexus_file_browser(root, config.get('default_directory')),
            'Jupyter Browser': lambda: create_jupyter_browser(root, widget.notebook_dir.get()),
            'Data Viewer': lambda: create_data_viewer(widget.data_dir.get(), root, config),
            'Range selector': lambda: create_range_selector(widget.data_dir.get(), root, config),
            'Log viewer': lambda: create_gda_terminal_log_viewer(widget.data_dir.get(), root),
            'Visit Viewer': lambda: create_visit_viewer(config),
        },
        'Processing': {
            'Script Editor': lambda: create_python_editor(None, root, config),
            'Open a terminal': lambda: open_terminal(f"cd {widget.data_dir.get()}"),
            'Start Jupyter (processing)': lambda: launch_jupyter_notebook('notebook', widget.proc_dir.get()),
            'Start Jupyter (notebooks)': lambda: launch_jupyter_notebook('notebook', widget.notebook_dir.get()),
            'Stop Jupyter servers': terminate_notebooks,
        }
    }
    menu.update(widget.menu_items())

    topmenu(root, menu, add_themes=True, add_about=True, config=config)

    root.mainloop()
    return root



"""
widget for selecting scan numbers and s
"""

import tkinter as tk
from tkinter import ttk
import asteval

from mmg_toolbox.utils.env_functions import (get_scan_numbers, scan_number_mapping)
from mmg_toolbox.utils.file_functions import get_scan_number
from ..misc.logging import create_logger
from ..misc.styles import create_hover
from ..misc.config import get_config, C
from .find_scans import FindScans

logger = create_logger(__file__)


class ScanRangeSelector:
    """Frame with """

    def __init__(self, root: tk.Misc, initial_directory: str | None = None, config: dict | None = None,
                 metadata_getter: tk.StringVar | None = None):
        logger.info('Creating ScanRangeSelector')
        self.root = root
        self.config = config or get_config()
        self.metadata = metadata_getter

        # variables
        self.exp_folder = tk.StringVar(self.root, initial_directory)
        self.number_start = tk.StringVar(self.root, '-10')
        self.number_end = tk.StringVar(self.root, '-1')
        self.number_step = tk.IntVar(self.root, 1)
        self.file_list = []

        # Range selection
        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

        ttk.Label(frm, text='First:').pack(side=tk.LEFT, padx=2)
        var = ttk.Entry(frm, textvariable=self.number_start, width=8)
        var.pack(side=tk.LEFT, padx=2)
        var.bind("<Return>", self.update_numbers)
        ttk.Label(frm, text='Last:').pack(side=tk.LEFT, padx=2)
        var = ttk.Entry(frm, textvariable=self.number_end, width=8)
        var.pack(side=tk.LEFT, padx=2)
        var.bind("<Return>", self.update_numbers)
        ttk.Label(frm, text='Step:').pack(side=tk.LEFT, padx=2)
        var = ttk.Entry(frm, textvariable=self.number_step, width=4)
        var.pack(side=tk.LEFT, padx=2)
        var.bind("<Return>", self.update_numbers)
        ttk.Button(frm, text='Get numbers', command=self.numbers_from_exp).pack(side=tk.LEFT)
        ttk.Button(frm, text='Generate', command=self.update_numbers).pack(side=tk.LEFT, padx=4)
        ttk.Button(frm, text='Select Files', command=self.select_files).pack(side=tk.LEFT, padx=4)
        ttk.Button(frm, text='Find', command=self.find_scans).pack(side=tk.LEFT, padx=4)

        # Text box
        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        ttk.Label(frm, text='Scans = ').pack(side=tk.LEFT, padx=2)
        self.text = tk.Text(frm, wrap=tk.WORD, width=65, height=5)
        self.text.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)

        var = ttk.Scrollbar(frm, orient=tk.VERTICAL, command=self.text.yview)
        var.pack(side=tk.LEFT, fill=tk.Y)
        self.text.configure(yscrollcommand=var.set)
        ttk.Button(frm, text='Check', command=self.show_metadata).pack(side=tk.LEFT, fill=tk.Y)

    def numbers_from_exp(self):
        exp_folder = self.exp_folder.get()
        if exp_folder:
            numbers = get_scan_numbers(exp_folder)
            self.number_start.set(str(numbers[0]))
            self.number_end.set(str(numbers[-1]))

    def update_numbers(self, event=None):
        first = eval(self.number_start.get())
        last = eval(self.number_end.get())
        step = self.number_step.get()

        if (last - first) / step > 1000:
            raise IOError('Range is too large')

        exp_folder = self.exp_folder.get()
        if not exp_folder:
            return
        scan_numbers = get_scan_numbers(exp_folder)
        if first < 1 or last < 1:
            last_scan = scan_numbers[-1]
            if first < 1:
                first = last_scan + first
            if last < 1:
                last = last_scan + last

        scan_range = [n for n in range(first, last+1, step) if n in scan_numbers]
        self.text.replace("1.0", tk.END, str(scan_range))

    def generate_scan_numbers(self) -> list[int]:
        scan_text = self.text.get("1.0", tk.END)
        if not scan_text.strip():
            return []
        safe_eval = asteval.Interpreter(use_numpy=True)
        scan_numbers = safe_eval(scan_text)
        return scan_numbers

    def generate_scan_files(self) -> dict[int, str]:
        scan_numbers = self.generate_scan_numbers()
        scan_files = scan_number_mapping(self.exp_folder.get())
        all_scan_numbers = list(scan_files)
        scan_numbers = [all_scan_numbers[n] if n < 1 else n for n in scan_numbers]
        return {number: scan_files[number] for number in scan_numbers if number in scan_files}

    def select_files(self):
        from ..apps.scans import list_scans
        file_list = list_scans(self.exp_folder.get())
        files = list_scans(*file_list, parent=self.root, config=self.config, button_name='Select')
        if files:
            numbers = [get_scan_number(file) for file in files]
            self.text.replace("1.0", tk.END, str(numbers))

    def show_metadata(self):
        from ..apps.scans import list_scans
        file_list = self.generate_scan_files().values()
        if len(file_list) == 0:
            return
        metadata_list = self.metadata.get().split(',') if self.metadata.get() else None
        list_scans(*file_list, parent=self.root, config=self.config, metadata_list=metadata_list)

    def find_scans(self):
        """Open scan finder widget"""

        scan_files = self.generate_scan_files()
        first_file = tuple(scan_files.values())[0] if scan_files else None
        metadata_list = self.metadata.get().split(',')

        top = self.root.winfo_toplevel()
        frame, fun_close = create_hover(top)
        widget = FindScans(frame, self.exp_folder.get(), self.config, first_file, metadata_list, close_fun=fun_close)
        scan_numbers = widget.wait_for_numbers()
        if scan_numbers:
            self.text.replace("1.0", tk.END, str(scan_numbers))



"""
tkinter search functions
"""

import tkinter as tk
from tkinter import ttk


def search_text(text: tk.Text, query="", match_case=False, highlight_colour='red'):
    """
    Set selection of strings in Text widget based on search query
    :param text: tk.Text
    :param query: str search query
    :param match_case: if False, select items even if the case doesn't match
    :param highlight_colour: colour of the highlighted text
    :return:
    """
    text.tag_remove('search', '1.0', tk.END)
    text.tag_config('search', background=highlight_colour)

    count = 0
    set_see = True
    if query:
        if query not in text.mark_names():
            text.mark_set(query, '1.0')
        show_idx = text.index(query)
        idx = '1.0'
        while idx:
            idx = text.search(query, idx, nocase=not match_case, stopindex=tk.END)
            if not idx:
                return count
            last_idx = '%s+%dc' % (idx, len(query))
            text.tag_add('search', idx, last_idx)
            if set_see and idx > show_idx:  # set view point at first or next query point
                text.mark_set(query, last_idx)
                text.see(idx)
                set_see = False
            count += 1
            idx = last_idx
    return count


def search_tree(treeview: ttk.Treeview, branch="", query="entry", match_case=False, whole_word=False):
    """
    Set selection of items in treeview based on search query
    :param treeview: ttk.treeview
    :param branch: ttk.treeview item (str)
    :param query: str search query
    :param match_case: if False, select items even if the case doesn't match
    :param whole_word: if True, select only items where query matches final element of address
    :return:
    """
    query = query if match_case else query.lower()
    for child in treeview.get_children(branch):
        search_tree(treeview, child, query, match_case, whole_word)
        address = treeview.item(child)['text']
        address = address if match_case else address.lower()
        address = address.split('/')[-1] if whole_word else address
        if (whole_word and query == address) or (not whole_word and query in address):
            treeview.selection_add(child)
            treeview.see(child)


import tkinter as tk

from mmg_toolbox.tkguis.misc.config import get_config
from mmg_toolbox.tkguis.misc.styles import RootWithStyle, create_root


def create_script_runner(parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    Create a range selector
    """
    from ..widgets.script_runner import ScriptRunner

    root = create_root(parent=parent, window_title='Script Runner')
    config = get_config() if config is None else config

    ScriptRunner(root, config)

    if parent is None:
        root.mainloop()
    return root



"""
lmfit models
"""

from lmfit import Model
from lmfit.models import (
    GaussianModel, LorentzianModel, VoigtModel, PseudoVoigtModel, LinearModel, ExponentialModel,
    SineModel
)

__all__ = ['PEAK_PARS', 'METHODS', 'get_peak_model', 'get_background_model', 'get_default_model',
           'PEAK_MODELS', 'BACKGROUND_MODELS']

# https://lmfit.github.io/lmfit-py/builtin_models.html#peak-like-models

ModelType = type[GaussianModel | LorentzianModel | VoigtModel | PseudoVoigtModel | LinearModel | ExponentialModel | SineModel]

MODELS: dict[str, type[ModelType]] = {
    'gaussian': GaussianModel,
    'lorentz': LorentzianModel,
    'voight': VoigtModel,
    'pvoight': PseudoVoigtModel,
    'linear': LinearModel,
    'exponential': ExponentialModel,
    'SineModel': SineModel,
}  # list of available lmfit models

PEAK_MODELS: dict[str, list[str]] = {
    'gaussian': ['gaussian', 'gauss'],
    'voight': ['voight', 'voight model'],
    'pvoight': ['pseudovoight', 'pvoight'],
    'lorentz': ['lorentz', 'lorentzian', 'lor'],
}  # alternative names for peaks

BACKGROUND_MODELS: dict[str, list[str]] = {
    'linear': ['flat', 'slope', 'linear', 'line', 'straight'],
    'exponential': ['exponential', 'curve']
}  # alternative names for background models

DEFAULT_MODEL = 'gaussian'
DEFAULT_BACKGROUND = 'linear'

# https://lmfit.github.io/lmfit-py/fitting.html#fit-methods-table
METHODS: dict[str, str] = {
    'leastsq': 'Levenberg-Marquardt',
    'nelder': 'Nelder-Mead',
    'lbfgsb': 'L-BFGS-B',
    'powell': 'Powell',
    'cg': 'Conjugate Gradient',
    'newton': 'Newton-CG',
    'cobyla': 'COBYLA',
    'bfgsb': 'BFGS',
    'tnc': 'Truncated Newton',
    'trust-ncg': 'Newton CG trust-region',
    'trust-exact': 'Exact trust-region',
    'trust-krylov': 'Newton GLTR trust-region',
    'trust-constr': 'Constrained trust-region',
    'dogleg': 'Dogleg',
    'slsqp': 'Sequential Linear Squares Programming',
    'differential_evolution': 'Differential Evolution',
    'brute': 'Brute force method',
    'basinhopping': 'Basinhopping',
    'ampgo': 'Adaptive Memory Programming for Global Optimization',
    'shgo': 'Simplicial Homology Global Ooptimization',
    'dual_annealing': 'Dual Annealing',
    'emcee': 'Maximum likelihood via Monte-Carlo Markov Chain',
}

# Peak parameter names
PEAK_PARS = ['amplitude', 'center', 'height', 'fwhm']


def get_peak_model(name: str) -> type[ModelType]:
    """Get peak model by name."""
    model: type[ModelType] | None = None
    for model_name, names in PEAK_MODELS.items():
        if name.lower() in names:
            model = MODELS[model_name]
    if model is None:
        raise ValueError(f"Peak model '{name}' not found")
    return model


def get_background_model(name: str) -> type[ModelType]:
    """Get background model by name."""
    model: type[ModelType] | None = None
    for model_name, names in BACKGROUND_MODELS.items():
        if name.lower() in names:
            model = MODELS[model_name]
    if model is None:
        raise ValueError(f"Background model '{name}' not found")
    return model


def get_default_model(peak: str = DEFAULT_MODEL, background: str = DEFAULT_BACKGROUND) -> Model:
    """Return a single peak lmfit model"""
    model = get_peak_model(peak)
    bkg = get_background_model(background)
    return model() + bkg()




"""
mmg_toolbox tests
Test utils.misc_functions
"""

import numpy as np
from mmg_toolbox.utils.misc_functions import (stfm, consolidate_numeric_strings, findranges, numbers2string,
                                              round_string_floats, shorten_string, data_holder)


def test_stfm():
    assert '35.25 (1)' == stfm(35.25, 0.01)
    assert '110 (5)' == stfm(110.25, 5)
    assert '0.0015300 (5)' == stfm(0.00153, 0.0000005)
    assert '1.56(2)E+6' == stfm(1.5632e6, 1.53e4)
    assert '0.00 (1)' == stfm(0, 0.01)


def test_consolidate_numeric_strings():
    my_range = list(range(12345, 12355)) + list(range(12357, 12365, 2))
    filenames = [f"i06-{n}.nxs" for n in my_range] + [f"i06-1-{n}.nxs" for n in my_range[:10]]
    strings = consolidate_numeric_strings(*filenames)
    assert len(strings) == 2
    assert strings == ['i06-####.nxs .. [123[45:54,57:2:63]]', 'i06-1-####.nxs .. [123[45:54]]']


def test_find_ranges():
    assert findranges([1, 2, 3, 4, 5]) == '1:5'
    assert findranges([1, 2, 3, 4, 5, 10, 12, 14, 16]) == '1:5,10:2:16'


def test_numbers2string():
    assert numbers2string([50001, 50002, 50003]) == '5000[1:3]'
    assert numbers2string([51020, 51030, 51040]) == '510[20:10:40]'


def test_round_string_floats():
    s = '#810002 scan eta 74.89533603616637 76.49533603616636 0.02 pil3_100k 1 roi2'
    assert round_string_floats(s) == '#810002 scan eta 74.895 76.495 0.02 pil3_100k 1 roi2'


def test_shorten_string():
    s = '\n #810002 scan eta 74.89533603616637 76.49533603616636 0.02 pol hkl checkbeam msmapper euler pil3_100k 1 roi2 \n other commands'
    assert shorten_string(s) == '#810002 scan eta 74.895 76.495 0.02 pol hkl checkbeam msmapper euler pil3_100k 1 roi2'
    out = shorten_string(s, 40, 10)
    assert len(out) == 40
    assert out == '#810002 scan eta 74.895 7 ... 00k 1 roi2'


def test_data_holder():
    scans = {'eta': np.arange(10), 'sum': np.arange(10)}
    metadata = {'a': 10, 'b': 20, 'c': 'hello'}
    d = data_holder(scans, metadata)
    assert hasattr(d, 'eta')
    assert hasattr(d, 'sum')
    assert hasattr(d.metadata, 'a')
    assert d.metadata.c == 'hello'





"""
Magnetic Materials Group Toolbox
"""

import sys
from mmg_toolbox.utils.file_reader import data_file_reader
from mmg_toolbox.utils.experiment import Experiment
from mmg_toolbox.beamline_metadata import metadata, xas_metadata, nexus_metadata

__version__ = '0.5.1'
__date__ = '20/01/2026'
__author__ = 'Dan Porter'

__all__ = ['start_gui', 'version_info', 'title', 'module_info',
           'data_file_reader', 'Experiment', 'metadata', 'xas_metadata', 'nexus_metadata']


def start_gui(*args: str):
    from mmg_toolbox.tkguis import run
    run(*args)


def version_info():
    return 'mmg_toolbox version %s (%s)' % (__version__, __date__)


def title():
    return 'mmg_toolbox  version %s' % __version__


def module_info():
    out = 'Python version %s' % sys.version
    out += '\n%s' % version_info()
    # Modules
    import numpy
    out += '\n     numpy version: %s' % numpy.__version__
    try:
        import matplotlib
        out += '\nmatplotlib version: %s' % matplotlib.__version__
    except ImportError:
        out += '\nmatplotlib version: None'
    try:
        import hdfmap
        out += '\nhdfmap version: %s (%s)' % (hdfmap.__version__, hdfmap.__date__)
    except ImportError:
        out += '\nhdfmap version: Not available'
    try:
        import tkinter
        out += '\n   tkinter version: %s' % tkinter.TkVersion
    except ImportError:
        out += '\ntkinter version: Not available'
    out += '\n'
    return out



"""
Automatic data file reader
"""

from datetime import datetime
from collections import defaultdict

from mmg_toolbox.utils.misc_functions import DataHolder
from mmg_toolbox.utils.dat_file_reader import read_dat_file
from mmg_toolbox.utils.env_functions import get_beamline
from mmg_toolbox.nexus.nexus_reader import read_nexus_file, NexusDataHolder


def data_file_reader(filename: str, beamline: str | None = None) -> NexusDataHolder | DataHolder:
    """
    Read Nexus or dat file as DataHolder
    """
    beamline = beamline or get_beamline(None, filename=filename)
    if filename.endswith('.dat'):
        return read_dat_file(filename)
    return read_nexus_file(filename, beamline=beamline)


def read_gda_terminal_log(filename: str) -> dict[str, list[str]]:
    """Read GDA terminal log using specific time stamp regex"""
    dt_format = '%Y-%m-%d %H:%M:%S,%f'
    line2dt = lambda ln: datetime.strptime(ln.split('|')[0].strip(), dt_format)
    tab_title = '%a %d%b'

    tabs = defaultdict(list)
    with open(filename) as file:
        for line in file:
            try:
                time = line2dt(line)
                title = time.strftime(tab_title)
                tabs[title] += [line.strip()]
            except ValueError as e:
                pass
    return tabs



import tkinter as tk

from mmg_toolbox.tkguis.misc.config import get_config
from mmg_toolbox.tkguis.misc.styles import RootWithStyle, create_root


def create_multi_scan_analysis(parent: tk.Misc | None = None, config: dict | None = None,
                               exp_directory: str | None = None, proc_directory: str | None = None,
                               scan_numbers: list[int] | None = None, metadata: str | None = None,
                               x_axis: str | None = None, y_axis: str | None = None) -> RootWithStyle:
    """
    Create a range selector
    """
    from ..widgets.multi_scan_analysis import MultiScanAnalysis

    root = create_root(parent=parent, window_title='Multi-Scan Analysis')
    config = get_config() if config is None else config

    MultiScanAnalysis(root, config, exp_directory=exp_directory,
                      proc_directory=proc_directory, scan_numbers=scan_numbers,
                      metadata=metadata, x_axis=x_axis, y_axis=y_axis)

    if parent is None:
        root.mainloop()
    return root



"""
Functions for writing Nexus files
"""

import h5py
import numpy as np
import datetime
import json

from hdfmap.nexus import default_nxentry

import mmg_toolbox.nexus.nexus_names as nn
from mmg_toolbox.utils.file_functions import get_scan_number
from mmg_toolbox.nexus.nexus_transformations import TransformationAxis, RotationAxis, TranslationAxis, get_depends_on
from mmg_toolbox.utils.polarisation import polarisation_label_to_stokes, analyser_jones_matrix
from mmg_toolbox.utils.xray_utils import photon_wavelength


def add_nxclass(root: h5py.Group, name: str, nx_class: str, **attrs) -> h5py.Group:
    """Create NXclass group"""
    group = root.create_group(name, track_order=True)
    group.attrs[nn.NX_CLASS] = nx_class
    group.attrs.update(attrs)
    return group


def add_nxfield(root: h5py.Group, name: str, data,
                add_to_axes: bool = False, add_to_signal: bool = False,
                **attrs) -> h5py.Dataset:
    """Create NXfield for storing data"""
    field = root.create_dataset(name, data=data)
    field.attrs.update(attrs)
    if add_to_axes:
        prev_axes = list(root.attrs.get(nn.NX_AXES, []))
        root.attrs[nn.NX_AXES] = prev_axes + [name]
        root.attrs[nn.NX_INDICES.format(name)] = [len(prev_axes)]
    if add_to_signal and nn.NX_SIGNAL in root.attrs:
        if nn.NX_AUXILIARY in root.attrs:
            root.attrs[nn.NX_AUXILIARY] = list(root.attrs[nn.NX_AUXILIARY]) +[name]
        else:
            root.attrs[nn.NX_AUXILIARY] = [name]
    elif add_to_signal:
        root.attrs[nn.NX_SIGNAL] = name
    return field


def add_attr(root: h5py.Group | h5py.Dataset, **attrs):
    """Add attributes to NXclass or NXfield"""
    root.attrs.update(attrs)


def add_nxentry(root: h5py.File, name: str, definition: str | None = None, default: bool = False) -> h5py.Group:
    """Create NXentry group"""
    entry = add_nxclass(root, name, nn.NX_ENTRY)
    if definition is not None:
        add_nxfield(entry, nn.NX_DEFINITION, definition)
    if default:
        root.attrs[nn.NX_DEFAULT] = name
    return entry


def add_nxinstrument(root: h5py.Group, name: str, instrument_name: str) -> h5py.Group:
    """Create NXinstrument group"""
    instrument = add_nxclass(root, name, nn.NX_INST)
    add_nxfield(instrument, nn.NX_NAME, instrument_name, short_name=instrument_name)
    return instrument


def add_nxsource(root: h5py.Group, name: str, source_name: str = 'dls', source_type: str = 'Synchrotron X-ray Source',
                 probe: str = 'x-ray', energy_gev: float = 3.0) -> h5py.Group:
    """
    Create NXsource group for DLS
    """
    source = add_nxclass(root, name, nn.NX_SOURCE)
    dls = 'Diamond Light Source'
    add_nxfield(source, nn.NX_NAME, dls if source_name.lower() == 'dls' else source_name,
                short_name=dls)
    add_nxfield(source, 'type', source_type)
    add_nxfield(source, 'probe', probe)
    add_nxfield(source, 'energy', energy_gev, units='GeV')
    return source


def add_nxinsertion_device(root: h5py.Group, name: str, id_type: str = 'undulator', gap: float = 7.0,
                           harmonic: int = 7) -> h5py.Group:
    """
    Create NXinsertion_device group
    """
    group = add_nxclass(root, name, nn.NX_ID)
    add_nxfield(group, 'id_type', id_type)
    add_nxfield(group, 'gap', gap, units='mm')
    add_nxfield(group, 'harmonic', harmonic)
    return group


def add_nxmono(root: h5py.Group, name: str, energy_ev: float | np.ndarray, units='eV') -> h5py.Group:
    """
    Create NXmonochromator group
    """
    mono = add_nxclass(root, name, nn.NX_MONO)
    add_nxfield(mono,'energy', energy_ev, units=units)
    return mono


def add_nxcrystal(root: h5py.Group, name: str, usage: str, crystal_type: str, d_spacing: float,
                  reflection: tuple[int, int, int] = (1, 1, 1), order_no: int = 1) -> h5py.Group:
    """
    Create NXcrystal group
    Used for monochromators and analyser crystals

    :param root: h5py group
    :param name: name of crystal
    :param usage: 'Bragg' or 'Laue'
    :param crystal_type: Chemical formula or substance, e.g. Si, HOPG, multilayer
    :param d_spacing: spacing between crystals in Angstrom
    :param reflection: reflection index in Miller-indices (h,k,l)
    :param order_no: order of the reflection, n*(h,k,l)
    :return: NXcrystal group
    """
    crystal = add_nxclass(root, name, nn.NX_XTL)
    add_nxfield(crystal, 'usage', usage)
    add_nxfield(crystal, 'type', crystal_type)
    add_nxfield(crystal, 'd_spacing', d_spacing, units='angstrom')
    add_nxfield(crystal, 'reflection', np.array(reflection, dtype=int).reshape(3))
    add_nxfield(crystal,'order_no', int(order_no))
    return crystal


def add_nxdetector(root: h5py.Group, name: str, data: np.ndarray,
                   detector_type: str = 'ccd',
                   detector_distance_mm: float = 1000, pixel_size_mm: float = 0.055,
                   depends_on: str | h5py.Group | None = None,
                   transformations: list[TransformationAxis] | None = None) -> h5py.Group:
    """
    Create NXdetector group
    """
    detector = add_nxclass(root, name, nn.NX_DET)
    add_nxfield(detector, 'data', data)
    add_nxfield(detector, 'type', detector_type)

    if transformations is None:
        transformations = (TranslationAxis('detector_offset', detector_distance_mm, vector=(0, 0, 1)),)
    add_nxtransformations(detector, 'transformations', *transformations, depends_on=depends_on)
    path = detector[nn.NX_DEPON][...]

    # Add module directions if 2D detector
    if data.ndim >= 3:
        module = add_nxclass(detector, 'module', nn.NX_MODULE)
        add_nxfield(module, nn.NX_MODULE_ORIGIN, (0, 0))
        add_nxfield(module, nn.NX_MODULE_SIZE, data.shape[-2:])
        add_nxfield(module, nn.NX_MODULE_OFFSET, 0, units='mm', transformation_type=nn.NX_TTRAN,
                    vector=(0, 0, 0), offset=(0, 0, 0), offset_units='mm', depends_on=path)
        add_nxfield(module, nn.NX_MODULE_FAST, pixel_size_mm, units='mm', transformation_type=nn.NX_TTRAN,
                    vector=(0, 0, 0), offset=(0, 0, 0), offset_units='mm', depends_on=path)
        add_nxfield(module, nn.NX_MODULE_SLOW, pixel_size_mm, units='mm', transformation_type=nn.NX_TTRAN,
                    vector=(0, 0, 0), offset=(0, 0, 0), offset_units='mm', depends_on=path)
    return detector


def add_nxbeam(root: h5py.Group, name: str, incident_energy_ev: float, polarisation_label: str = 'lh',
               beam_size_um: tuple[float, float] | None = None, arbitrary_polarisation_angle: float = 0.0) -> h5py.Group:
    """Create NXbeam group"""
    beam = add_nxclass(root, name, nn.NX_BEAM)
    # Fields
    add_nxfield(beam, nn.NX_EN, incident_energy_ev, units='eV')
    wl = photon_wavelength(incident_energy_ev / 1000.)
    add_nxfield(beam, nn.NX_WL, wl, units='angstrom')
    pol_stokes = polarisation_label_to_stokes(polarisation_label, arbitrary_polarisation_angle)
    add_nxfield(beam, nn.NX_STOKES, pol_stokes)
    if beam_size_um is not None:
        add_nxfield(beam, nn.NX_BSIZE, beam_size_um, units='μm')
    return beam


def add_nxsample(root: h5py.Group, name: str, sample_name: str = '', chemical_formula: str = '',
                 temperature_k: float = 300, magnetic_field_t: float = 0, electric_field_v: float = 0,
                 mag_field_dir: str = 'z', electric_field_dir: str = 'z',
                 sample_type: str = 'sample', description: str = '') -> h5py.Group:
    """Create NXsample group"""
    sample = add_nxclass(root, name, nn.NX_SAMPLE)
    # fields
    add_nxfield(sample, 'name', sample_name)
    add_nxfield(sample, 'chemical_formula', chemical_formula)
    add_nxfield(sample, 'type', sample_type)
    add_nxfield(sample, 'description', description)
    add_nxfield(sample, 'temperature', temperature_k, units='K')
    add_nxfield(sample, 'magnetic_field', magnetic_field_t, units='T', direction=mag_field_dir)
    add_nxfield(sample, 'electric_field', electric_field_v, units='V', direction=electric_field_dir)
    return sample


def add_nxdata(root: h5py.Group, name: str, axes: list[str], signal: str, *auxilliary_signals: str,
               default: bool = False) -> h5py.Group:
    """
    Create NXdata group

    xvals = np.arange(10)
    yvals = 3 + xvals ** 2
    group = NXdata(entry, 'xydata', axes=['x'], signal='y')
    xdata = add_nxfield(group, 'x', xvals, units='mm')
    ydata = add_nxfield(group, 'y' yvals, units='')
    """
    group = add_nxclass(root, name, nn.NX_DATA)
    group.attrs.update({
        nn.NX_AXES: list(axes),
        nn.NX_SIGNAL: signal,
        nn.NX_AUXILIARY: list(auxilliary_signals),
    })
    group.attrs.update({
        nn.NX_INDICES.format(ax_name): [n] for n, ax_name in enumerate(axes)
    })
    if default:
        root.attrs['default'] = name
    return group


def add_nxmonitor(root: h5py.Group, name: str, data: np.ndarray | str) -> h5py.Group:
    """
    Create NXmonitor group with monitor signal
    """
    group = add_nxclass(root, name, nn.NX_MON)
    if isinstance(data, str):
        group['data'] = h5py.SoftLink(data)
    else:
        group.create_dataset('data', data=data)
    return group


def add_nxnote(root: h5py.Group, name: str, description: str, data: str | dict | None = None,
               filename: str | None = None, sequence_index: int | None = None) -> h5py.Group:
    """
    add NXnote to parent group
    """
    note = add_nxclass(root, name, nn.NX_NOTE)
    if isinstance(data, dict):
        note.create_dataset('type', data='application/json')
    else:
        note.create_dataset('type', data='text/html')
    note.create_dataset('description', data=str(description))
    if filename:
        note.create_dataset('file_name', data=str(filename))
    if data:
        if isinstance(data, dict):
            note.create_dataset('data', data=json.dumps(data))
        else:
            note.create_dataset('data', data=data.encode('utf-8'))
    if sequence_index:
        note.create_dataset('sequence_index', data=int(sequence_index))
    return note


def add_nxparameters(root: h5py.Group, name: str, **parameters) -> h5py.Group:
    """
    Add NXparameters to parent group
    """
    group = add_nxclass(root, name, nn.NX_PARAM)
    for key, value in parameters.items():
        attrs = {}
        if isinstance(value, tuple):
            value, units = value
            attrs['units'] = units
        value = np.array(value)
        if not np.issubdtype(value.dtype, np.number):
            value = value.astype(bytes)  # convert None and other obj to str
        add_nxfield(group, key, data=value, **attrs)
    return group


def add_nxprocess(root: h5py.Group, name: str, program: str | None = None,
                  version: str | None = None, date: str | None = None, sequence_index: int | None = None,
                  **parameters) -> h5py.Group:
    """
    Create NXprocess group

    Example:
    entry = add_nxentry(root, 'processed')
    process = add_nxprocess(entry, 'process', program='Python', version='1.0')
    add_nxnote(process, 'step_1',
            description='First step',
            data='details',
            sequence_index=1
    )
    add_nxnote(process, 'step_2',
            description='Second step',
            data='details',
            sequence_index=2
    )
    data = add_nxdata(process, 'result', axes=['x'], signal='y')
    xdata = add_nxfield(group, 'x', xvals, units='mm')
    ydata = add_nxfield(group, 'y' yvals, units='')
    """
    from mmg_toolbox import version_info

    if program is None:
        program = 'Python:mmg_toolbox'
    if version_info is None:
        version = version_info()
    if date is None:
        date = str(datetime.datetime.now())

    group = add_nxclass(root, name, nn.NX_PROC)
    if sequence_index:
        group.create_dataset('sequence_index', data=int(sequence_index))
    group.create_dataset('program', data=str(program))
    group.create_dataset('version', data=str(version))
    group.create_dataset('date', data=str(date))
    if parameters:
        add_nxparameters(group, 'parameters', **parameters)
    return group


def add_entry_links(root: h5py.File, *filenames: str):
    """
    Add entry links to nexus file
    """
    # Add links to previous files
    for n, filename in enumerate(filenames):
        if not h5py.is_hdf5(filename):
            continue
        with h5py.File(filename) as nxs:
            entry_path = default_nxentry(nxs)
        number = get_scan_number(filename) or n + 1
        label = str(number)
        root[label] = h5py.ExternalLink(filename, entry_path)


def add_channel_cut_mono(instrument: h5py.Group, name: str, energy: float | np.ndarray,
                         d_spacing: float, crystal_type: str = 'Si', units: str = 'eV',
                         reflection: tuple[int, int, int] = (1, 1, 1), order_no: int = 1):
    """
    Add channel cut mono
    """
    # create mono
    mono = add_nxmono(instrument, name, energy, units)
    # Add crystals
    add_nxcrystal(
        root=mono,
        name='crystal1',
        usage='Bragg',
        d_spacing=d_spacing,
        crystal_type=crystal_type,
        reflection=reflection,
        order_no=order_no
    )
    add_nxcrystal(
        root=mono,
        name='crystal2',
        usage='Bragg',
        d_spacing=d_spacing,
        crystal_type=crystal_type,
        reflection=reflection,
        order_no=order_no
    )


def add_analyser_detector(instrument: h5py.Group, name: str, data: np.ndarray,
                          d_spacing: float, crystal_type: str = 'HOPG',
                          reflection: tuple[int, int, int] = (0, 0, 1), order_no: int = 1,
                          bragg: float = 90, stokes: float = 0,
                          sample_analyser_distance_mm: float = 1000, analyser_det_distance_mm: float = 50,
                          pixel_size_mm: float = 0.1, depends_on: str | h5py.Group = '.'
                          ):
    """
    Add analyser detector
    """
    xtl = add_nxcrystal(
        root=instrument,
        name=f"{name}_analyser",
        usage='Bragg',
        d_spacing=d_spacing,
        crystal_type=crystal_type,
        reflection=reflection,
        order_no=order_no,
    )

    # Transformations - Analyser
    ana_trans = TranslationAxis(name='origin', value=sample_analyser_distance_mm, vector=(0, 0, 1))
    rot_stokes = RotationAxis(name='Stokes', value=stokes, vector=(0, 0, 1))
    add_nxtransformations(xtl, 'transformations', ana_trans, rot_stokes, depends_on=depends_on)

    # Transformations - Detector
    det_trans = TranslationAxis(name='origin_offset', value=analyser_det_distance_mm, vector=(0, 0, 1))
    det_bragg = RotationAxis(name='tthp', value=bragg, vector=(-1, 0, 0))
    # Add detector
    detector = add_nxdetector(
        root=instrument,
        name=name,
        data=data,
        detector_distance_mm=analyser_det_distance_mm,
        pixel_size_mm=pixel_size_mm,
        depends_on=xtl,
        transformations=[det_bragg, det_trans],
    )
    # Add Jones matrix
    jones = analyser_jones_matrix(crystal_bragg=bragg, rotation=stokes)
    add_nxfield(detector, 'polarization_analyser_jones_matrix', jones)


def add_6circle_diffractometer(instrument: h5py.Group, name: str, phi: np.ndarray, chi: np.ndarray, eta: np.ndarray,
                               mu: np.ndarray, delta: np.ndarray, gamma: np.ndarray) -> h5py.Group:
    """6-circle Euler diffractometer"""
    from mmg_toolbox.diffraction.diffcalc import euler2kappa, KALPHA
    kphi, kappa, ktheta = euler2kappa(phi, chi, eta, mode=1, kalpha=KALPHA)

    # Positions
    diff = add_nxclass(instrument, name, 'NXcollection')
    add_nxfield(diff, 'kphi', kphi, units='degrees')
    add_nxfield(diff, 'kappa', kappa, units='degrees')
    add_nxfield(diff, 'ktheta', ktheta, units='degrees')
    add_nxfield(diff, 'kalpha', KALPHA, units='degrees')
    add_nxfield(diff, 'phi', phi, units='degrees')
    add_nxfield(diff, 'chi', chi, units='degrees')
    add_nxfield(diff, 'eta', eta, units='degrees')
    add_nxfield(diff, 'mu', mu, units='degrees')
    add_nxfield(diff, 'delta', delta, units='degrees')
    add_nxfield(diff, 'gamma', gamma, units='degrees')

    # Transformations
    phi = RotationAxis('phi', phi, vector=(0, 1, 0))
    chi = RotationAxis('chi', chi, vector=(0, 0, 1))
    eta = RotationAxis('eta', eta, vector=(-1, 0, 0))
    mu = RotationAxis('mu', mu, vector=(0, 1, 0))
    delta = RotationAxis('delta', delta, vector=(-1, 0, 0))
    gamma = RotationAxis('gamma', gamma, vector=(0, 1, 0))

    add_nxtransformations(diff, 'sample', phi, chi, eta, mu, delta, gamma)
    del diff['depends_on']
    add_nxtransformations(diff, 'detector_arm', delta, gamma)
    del diff['depends_on']
    return diff


def add_6circle_diffractometer_kappa(instrument: h5py.Group, name: str,
                                     kphi: np.ndarray, kappa: np.ndarray, ktheta: np.ndarray,
                                     mu: np.ndarray, delta: np.ndarray, gamma: np.ndarray,
                                     kalpha: float,) -> h5py.Group:
    """6-circle Kappa diffractometer"""
    from mmg_toolbox.diffraction.diffcalc import kappa2euler
    phi, chi, eta = kappa2euler(ktheta, kappa, kphi, mode=1, kalpha=kalpha)

    # Positions
    diff = add_nxclass(instrument, name, 'NXcollection')
    add_nxfield(diff, 'kphi', kphi, units='degrees')
    add_nxfield(diff, 'kappa', kappa, units='degrees')
    add_nxfield(diff, 'ktheta', ktheta, units='degrees')
    add_nxfield(diff, 'kalpha', kalpha, units='degrees')
    add_nxfield(diff, 'phi', phi, units='degrees')
    add_nxfield(diff, 'chi', chi, units='degrees')
    add_nxfield(diff, 'eta', eta, units='degrees')
    add_nxfield(diff, 'mu', mu, units='degrees')
    add_nxfield(diff, 'delta', delta, units='degrees')
    add_nxfield(diff, 'gamma', gamma, units='degrees')

    # Transformations
    phi = RotationAxis('phi', phi, vector=(0, 1, 0))
    chi = RotationAxis('chi', chi, vector=(0, 0, 1))
    eta = RotationAxis('eta', eta, vector=(-1, 0, 0))
    mu = RotationAxis('mu', mu, vector=(0, 1, 0))
    delta = RotationAxis('delta', delta, vector=(-1, 0, 0))
    gamma = RotationAxis('gamma', gamma, vector=(0, 1, 0))

    add_nxtransformations(diff, 'sample', phi, chi, eta, mu, delta, gamma)
    del diff['depends_on']
    add_nxtransformations(diff, 'detector_arm', delta, gamma)
    del diff['depends_on']
    return diff


############################################################
################### NXTransformations ######################
############################################################


def add_nxtransformations(root: h5py.Group, name: str, *transformations: TransformationAxis,
                          depends_on: str | h5py.Group | None = None) -> h5py.Group:
    """
    Add NXtransformations group
    Adds a depends_on dataset to the given group and a sub-group called 'transformations'
    """
    depends_on_ds = root.create_dataset(nn.NX_DEPON, data='.')
    if len(transformations) == 0:
        depends_on_ds[...] = get_depends_on(depends_on)
        return root
    group = add_nxclass(root, name, nn.NX_TRAN)
    axes = []
    for t in transformations:
        axis = add_nxfield(
            root=group,
            name=t.name,
            data=t.value,
            units=t.units,
            transformation_type=t.type,
            vector=t.vector,
            offset=t.offset,
            offset_units=t.offset_units,
            depends_on='.'
        )
        axes.append(axis)

    # Assign depends on pointers
    for axis, do in zip(axes, axes[1:]):
        axis.attrs[nn.NX_DEPON] = do.name
    # Point group depends_on dataset to the first transformation in the chain
    depends_on_ds[...] = axes[0].name
    # Point the last transformation towards an external transformation dataset
    if depends_on is not None:
        if isinstance(depends_on, h5py.Group):
            if nn.NX_DEPON in depends_on:
                depends_on = depends_on[nn.NX_DEPON].asstr()[...]
            else:
                raise Exception(f"group: {depends_on.name} does not contain 'depends_on'")
        axes[-1].attrs[nn.NX_DEPON] = depends_on  # path of transformation dataset
    return group


